import numpy as np
import scipy as scp
import pandas as pd
from mssm.models import *
from mssm.src.python.smooths import convolve_event
from mssm.src.python.gamm_solvers import cpp_cholP,compute_Linv,apply_eigen_perm
from mssm.src.python.formula import reparam

################################## Contains simulations to simulate for GAMM & GAMMLSS models ##################################

def sim1(sim_size,sim_sigma = 5.5,sim_lam = 1e-4,sim_weak_nonlin = 0.5,random_seed=None,fixed_seed=42*3):
    """First simulation for an additive time-series model with trial-level non-linear random effects. Data-set contains covariates time & x,
    for which the effect is different for three levels of factor variable fact.

    :param sim_size: Number of trials, defaults to 1000
    :type sim_size: int, optional
    :param sim_sigma: Standard error for residuals, defaults to 5.5
    :type sim_sigma: float, optional
    :param sim_lam: Lambda parameter for trial-level non-linear effect complexity, defaults to 1e-4
    :type sim_lam: _type_, optional
    :param sim_weak_nonlin: Strength of weakly non-linear covariate x effect, defaults to 0.5
    :type sim_weak_nonlin: float, optional
    :param random_seed: Seed for random parts of the simulation - should differ between repeated simulations, defaults to None
    :type random_seed: int, optional
    :param fixed_seed: Seed for fixed effects in the simulation - should NOT differ between repeated simulations, defaults to None
    :type fixed_seed: int, optional
    :return: Tuple, first element contains a ``pd.DataFrame`` with simulated data, second element is again a tuple containing: a ``np.array`` with the trial-level deviations, design matrices used for simulation, true coefficients used for simulation, and true intercepts used for simulation.
    :rtype: (pd.Dataframe,(np.array,np.array,np.array,np.array,np.array,np.array))
    """
    global i
    global f
    global lhs


    # Set up fixed and random effects
    time_pred = np.array([t for t in range(0,3000,20)])
    x_pred = np.linspace(0,25,len(time_pred))

    # Get matrix for time effects
    sim_dat = pd.DataFrame({"Time":time_pred,
                            "x":x_pred,
                            "y":scp.stats.norm.rvs(size=len(time_pred),random_state=20)})

    sim_formula = Formula(lhs("y"),
                        [i(),f(["Time"],nk=15)],
                        data=sim_dat,
                        print_warn=False)

    sim_model = GAMM(sim_formula,Gaussian())
    sim_model.fit(progress_bar=False)
    sim_mat = sim_model.get_mmat()

    sim_S = sim_formula.penalties[0].S_J_emb * sim_lam

    # Get fixed time effects (+intercept)
    fixed1 = np.array([5,*scp.stats.norm.rvs(size=(sim_S.shape[1]-1),scale=5,random_state=fixed_seed)]).reshape(-1,1)
    fixed2 = np.array([-5,*scp.stats.norm.rvs(size=(sim_S.shape[1]-1),scale=5,random_state=int(fixed_seed*3))]).reshape(-1,1)
    fixed3 = np.zeros_like(fixed2)
    fixed_sim_time_coefs = [fixed1,fixed2,fixed3]

    # Also get intercepts alone
    true_offsets = [fixed1[0],fixed2[0],fixed3[0]]

    # Prepare random smooth sampler
    # Based on Wood (2017, 6.10)
    nH = sim_mat.T @ sim_mat + sim_S
    Lp, Pr, code = cpp_cholP(nH.tocsc())

    if code != 0:
        raise ValueError("Cholesky failed.")
    
    LVp = compute_Linv(Lp,1)
    LV = apply_eigen_perm(Pr,LVp)
    V = (LV.T @ LV) * sim_sigma
    V = V.toarray()
    V[np.abs(V) < 1e-4] = 0 # Minimize numerical inaccuracies between different systems to ensure similar behavior

    #V = scp.sparse.linalg.spsolve(sim_mat.T @ sim_mat + sim_S,scp.sparse.eye((sim_S.shape[1]),format='csc')) * sim_sigma

    # Get matrix for x effects
    sim_formula2 = Formula(lhs("y"),
                        [i(),f(["x"],nk=5)],
                        data=sim_dat,
                        print_warn=False)

    sim_model2 = GAMM(sim_formula2,Gaussian())
    sim_model2.fit(progress_bar=False)
    sim_mat2 = sim_model2.get_mmat()

    # Get fixed x effects
    fixedX1 = np.array([0,*scp.stats.norm.rvs(size=(5),scale=5,random_state=fixed_seed*6)]).reshape(-1,1)
    fixedX2 = np.array([0,*scp.stats.norm.rvs(size=(5),scale=5,random_state=fixed_seed*9)]).reshape(-1,1)
    fixedX3 = np.array([0,*np.linspace(-sim_weak_nonlin,sim_weak_nonlin,len(fixedX2)-1)]).reshape(-1,1)
    fixed_sim_x_coefs = [fixedX1,fixedX2,fixedX3]
    
    ft = [] # series specific effect for each data point
    time = [] # time of each data point
    x = [] # x covariate of each data point
    il = [] # id of each data point
    fact = [] # group of each data point
    sub = [] # subject of each data point

    # Simulation seed
    np_gen = np.random.default_rng(random_seed)

    # Group assignment
    fl = np_gen.choice([1,2,3],size=sim_size,replace=True,p=[0.5,0.2,0.3])

    # x values for each trial
    xl = np_gen.choice(x_pred,size=sim_size,replace=True)

    # Sample trial-level smooths

    # random offsets
    rand_int = scp.stats.norm.rvs(size=sim_size,scale=2.5,random_state=random_seed)

    # random drifts
    rand_slope = scp.stats.norm.rvs(size=sim_size,scale=0.0025,random_state=random_seed)

    rand_matrix = np.zeros((100,len(time_pred)))
    for sim_idx in range(sim_size):
        if not random_seed is None:
            sample = scp.stats.multivariate_normal.rvs(mean=scp.stats.norm.rvs(size=(sim_S.shape[1]),scale=5,random_state=random_seed+sim_idx),cov=V,size=1,random_state=random_seed+sim_idx)
        else:
            sample = scp.stats.multivariate_normal.rvs(mean=scp.stats.norm.rvs(size=(sim_S.shape[1]),scale=5,random_state=None),cov=V,size=1,random_state=None)
        sample[0] = 0
        take = np_gen.integers(int(len(time_pred)/4),len(time_pred)+1)
        fact.extend(np.repeat(fl[sim_idx],take))
        sub.extend([f"sub_{sim_idx%20}" for _ in range(take)])
        time.extend(time_pred[0:take])
        x.extend(np.repeat(xl[sim_idx],take))
        il.extend(np.repeat(sim_idx,take))
        ft.extend(((sim_mat @ sample) + rand_int[sim_idx] + time_pred*rand_slope[sim_idx])[0:take])
        
        if sim_idx < rand_matrix.shape[0]:
            rand_matrix[sim_idx,:] += ((sim_mat @ sample) + rand_int[sim_idx] + time_pred*rand_slope[sim_idx])

    time = np.array(time)
    x = np.array(x)
    fact = np.array(fact)
    sub = np.array(sub)
    ft = np.array(ft).reshape(-1,1)

    # Get fixed predictions
    f0 = np.zeros((len(time))) # time
    f1 = np.zeros((len(time))) # x

    for fi in [1,2,3]:
        sim_cond_dat = pd.DataFrame({"Time":time[fact == fi]})
        sim_condX_dat = pd.DataFrame({"x":x[fact == fi]})
        _,sim_mat_cond,_ = sim_model.predict([0,1],sim_cond_dat)
        _,sim_matX_cond,_ = sim_model2.predict([0,1],sim_condX_dat)

        f0[fact == fi] = np.ndarray.flatten(sim_mat_cond@fixed_sim_time_coefs[fi-1])
        f1[fact == fi] = np.ndarray.flatten(sim_matX_cond@fixed_sim_x_coefs[fi-1])
    
    f0 = np.array(f0).reshape(-1,1)
    f1 = np.array(f1).reshape(-1,1)

    # Now build sim dat and define formula
    sim_fit_dat = pd.DataFrame({"y":np.ndarray.flatten(f0 + f1 + ft + scp.stats.norm.rvs(size=len(f0),scale=sim_sigma,random_state=random_seed).reshape(-1,1)),
                                "truth":np.ndarray.flatten(f0 + f1),
                                "time":time,
                                "x":x,
                                "fact":[f"fact_{fc}" for fc in fact],
                                "sub":sub,
                                "series":[f"series_{ic}" for ic in il]})

    return sim_fit_dat,(rand_matrix,sim_mat,sim_mat2,fixed_sim_time_coefs,fixed_sim_x_coefs,true_offsets)

def sim2(sim_size,sim_sigma = 5.5,sim_lam = 1e-4,set_zero = 1,random_seed=None,fixed_seed=42*3):
    """Second simulation for an additive time-series model with trial-level non-linear random effects. Data contains two
    additional covariates apart from time (x & z) - values for z vary within and between series, x only between series.

    Ground-truth for x or z can be set to zero.

    :param sim_size: Number of trials, defaults to 1000
    :type sim_size: int, optional
    :param sim_sigma: Standard error for residuals, defaults to 5.5
    :type sim_sigma: float, optional
    :param sim_lam: Lambda parameter for trial-level non-linear effect complexity, defaults to 1e-4
    :type sim_lam: _type_, optional
    :param set_zero: Which covariate (1 or 2 for x and z respectively) to set to zero, defaults to 1
    :type set_zero: int, optional
    :param random_seed: Seed for random parts of the simulation - should differ between repeated simulations, defaults to None
    :type random_seed: int, optional
    :param fixed_seed: Seed for fixed effects in the simulation - should NOT differ between repeated simulations, defaults to None
    :type fixed_seed: int, optional
    :return: Tuple, first element contains a ``pd.DataFrame`` with simulated data, second element is again a tuple containing: a ``np.array`` with the trial-level deviations, design matrices used for simulation, true effects used for simulation, and true offset used for simulation.
    :rtype: (pd.Dataframe,(np.array,np.array,np.array,np.array,np.array,np.array,np.array,float))
    """
    global i
    global f
    global lhs

    # Set up fixed and random effects
    time_pred = np.array([t for t in range(0,3000,20)])
    x_pred = np.linspace(0,25,len(time_pred))
    z_pred = np.linspace(-1,1,len(time_pred))

    # Get matrix for time effects
    sim_dat = pd.DataFrame({"Time":time_pred,
                            "x":x_pred,
                            "z":z_pred,
                            "y":scp.stats.norm.rvs(size=len(time_pred),random_state=20)})

    sim_formula = Formula(lhs("y"),
                        [i(),f(["Time"],nk=15)],
                        data=sim_dat,
                        print_warn=False)

    sim_model = GAMM(sim_formula,Gaussian())
    sim_model.fit(progress_bar=False)
    sim_mat = sim_model.get_mmat()

    sim_S = sim_formula.penalties[0].S_J_emb * sim_lam

    # Get fixed time effects
    fixed_time = np.array([5,*scp.stats.norm.rvs(size=(sim_S.shape[1]-1),scale=5,random_state=fixed_seed)]).reshape(-1,1)

    # Also get intercept alone
    true_offset = 5

    # Prepare random smooth sampler
    # Based on Wood (2017, 6.10)
    nH = sim_mat.T @ sim_mat + sim_S
    Lp, Pr, code = cpp_cholP(nH.tocsc())
    
    if code != 0:
        raise ValueError("Cholesky failed.")
    
    LVp = compute_Linv(Lp,1)
    LV = apply_eigen_perm(Pr,LVp)
    V = (LV.T @ LV) * sim_sigma
    V = V.toarray()
    V[np.abs(V) < 1e-4] = 0 # Minimize numerical inaccuracies between different systems to ensure similar behavior

    #V = scp.sparse.linalg.spsolve(sim_mat.T @ sim_mat + sim_S,scp.sparse.eye((sim_S.shape[1]),format='csc')) * sim_sigma

    # Get matrix for x effects
    sim_formula2 = Formula(lhs("y"),
                        [i(),f(["x"],nk=5)],
                        data=sim_dat,
                        print_warn=False)

    sim_model2 = GAMM(sim_formula2,Gaussian())
    sim_model2.fit(progress_bar=False)
    sim_mat2 = sim_model2.get_mmat()

    # Get fixed x effects
    fixed_x = np.array([0,*scp.stats.norm.rvs(size=(5),scale=5,random_state=int(fixed_seed*6))]).reshape(-1,1)

    # Get matrix for z effects
    sim_formula3 = Formula(lhs("y"),
                        [i(),f(["z"],nk=10)],
                        data=sim_dat,
                        print_warn=False)

    sim_model3 = GAMM(sim_formula3,Gaussian())
    sim_model3.fit(progress_bar=False)
    sim_mat3 = sim_model3.get_mmat()

    # Get fixed z effects
    fixed_z = np.array([0,*scp.stats.norm.rvs(size=(10),scale=5,random_state=int(fixed_seed*15))]).reshape(-1,1)

    # Simulation seed
    np_gen = np.random.default_rng(random_seed)
    
    ft = [] # series specific effect for each data point
    time = [] # time of each data point
    x = [] # x covariate of each data point
    il = [] # id of each data point

    # x values for each trial
    xl = np_gen.choice(x_pred,size=sim_size,replace=True)

    # Sample trial-level smooths

    # random offsets
    rand_int = scp.stats.norm.rvs(size=sim_size,scale=2.5,random_state=random_seed)
    
    # random drifts
    rand_slope = scp.stats.norm.rvs(size=sim_size,scale=0.0025,random_state=random_seed)
    
    rand_matrix = np.zeros((100,len(time_pred)))
    for sim_idx in range(sim_size):
        if not random_seed is None:
            sample = scp.stats.multivariate_normal.rvs(mean=scp.stats.norm.rvs(size=(sim_S.shape[1]),scale=5,random_state=random_seed+sim_idx),cov=V,size=1,random_state=random_seed+sim_idx)
        else:
            sample = scp.stats.multivariate_normal.rvs(mean=scp.stats.norm.rvs(size=(sim_S.shape[1]),scale=5,random_state=None),cov=V,size=1,random_state=None)
        sample[0] = 0
        take = np_gen.integers(int(len(time_pred)/4),len(time_pred)+1)
        
        time.extend(time_pred[0:take])
        x.extend(np.repeat(xl[sim_idx],take))
        il.extend(np.repeat(sim_idx,take))
        ft.extend(((sim_mat @ sample) + rand_int[sim_idx] + time_pred*rand_slope[sim_idx])[0:take])

        if sim_idx < rand_matrix.shape[0]:
            rand_matrix[sim_idx,:] + ((sim_mat @ sample) + rand_int[sim_idx] + time_pred*rand_slope[sim_idx])

    time = np.array(time)
    x = np.array(x)
    z = np_gen.choice(z_pred,size=len(time),replace=True) # z covariate of each data point
    ft = np.array(ft).reshape(-1,1)

    # Get fixed predictions
    sim_time_dat = pd.DataFrame({"Time":time})
    sim_X_dat = pd.DataFrame({"x":x})
    sim_Z_dat = pd.DataFrame({"z":z})

    _,sim_mat_time,_ = sim_model.predict([0,1],sim_time_dat)
    _,sim_mat_x,_ = sim_model2.predict([0,1],sim_X_dat)
    _,sim_mat_z,_ = sim_model3.predict([0,1],sim_Z_dat)

    f0 = sim_mat_time @ fixed_time # time
    f1 = sim_mat_x @ fixed_x # x
    f2 = sim_mat_z @ fixed_z # z

    # Set co-variate effects to zero
    if set_zero == 1:
        f1 = np.zeros_like(f1)

    if set_zero == 2:
        f2 = np.zeros_like(f2)

    # Now build sim dat and define formula
    sim_fit_dat = pd.DataFrame({"y":np.ndarray.flatten(f0 + f1 + f2 + ft + scp.stats.norm.rvs(size=len(f0),scale=sim_sigma,random_state=random_seed).reshape(-1,1)),
                                "truth":np.ndarray.flatten(f0 + f1 + f2),
                                "time":time,
                                "x":x,
                                "z":z,
                                "series":[f"series_{ic}" for ic in il]})
    return sim_fit_dat,(rand_matrix,sim_mat,sim_mat2,sim_mat3,fixed_time,fixed_x,fixed_z,true_offset)

def sim3(n,scale,c=1,binom_offset=0,family=Gaussian(),correlate=False,seed=None):
    """
    First Simulation performed by Wood et al., (2016): 4 smooths, 1 is really zero everywhere.
    Based on the original functions of Gu & Whaba (1991).

    This is also the first simulation performed by gamSim() - except for the fact that f(x_0) can also be set to
    zero, as was done by Wood et al., (2016)

    Covariates can also be simulated to correlate with each other, following the steps outlined in supplementary materials
    E of Wood et al., (2016).

    References:

     - Gu, C. & Whaba, G., (1991). Minimizing GCV/GML scores with multiple smoothing parameters via the Newton method.
     - Wood, S. N., Pya, N., Saefken, B., (2016). Smoothing Parameter and Model Selection for General Smooth Models
     - mgcv source code: gam.sim.r

    :param scale: Standard deviation for `family='Gaussian'` else scale parameter
    :type scale: float
    :param c: Effect strength for x3 effect - 0 = No effect, 1 = Maximal effect
    :type c: float
    :param binom_offset: Additive adjustment to log-predictor for Binomial model (5 in mgcv). Defaults to 0.
    :type binom_offset: float
    :param correlate: Whether predictor covariates should correlate or not. Defaults to False
    :type correlate: bool
    :param family: Distribution for response variable, must be: `Gaussian()`, `Gamma()`, or `Binomial()`. Defaults to `Gaussian()`
    :type family: Family, optional
    """
    np_gen = np.random.default_rng(seed)

    if correlate:
        # Following steps by Wood et al. (2016)
        Sigma = np.zeros((4,4)) + 0.9

        for ij in range(4):
            Sigma[ij,ij] = 1

        z = scp.stats.multivariate_normal.rvs(mean=[0 for _ in range(4)],cov=Sigma,size=n,random_state=seed)
        
        # I am a bit confused by notation in Wood et al. (2016) - they say x = cdf^{-1}(z) but I think that's not
        # what they mean, since cdf^{-1} = percent wise/quantile function which expects values between 0-1 which
        # is not the support for z. So I just use the cdf - which I think is what they mean. The resulting marginals
        # for x are uniform and all variables show the expected correlation, so it's probably correct.
        x0 = scp.stats.norm.cdf(z[:,0])
        x1 = scp.stats.norm.cdf(z[:,1])
        x2 = scp.stats.norm.cdf(z[:,2])
        x3 = scp.stats.norm.cdf(z[:,3])
    else:

        x0 = np_gen.random(n)
        x1 = np_gen.random(n)
        x2 = np_gen.random(n)
        x3 = np_gen.random(n)

    f0 = 2* np.sin(np.pi*x0)
    f1 = np.exp(2*x1)
    f2 = 0.2*np.power(x2,11)*np.power(10*(1-x2),6)+10*np.power(10*x2,3)*np.power(1-x2,10)
    f3 = np.zeros_like(x3)

    eta = c*f0 + f1 + f2 + f3 # eta in truth for non-Gaussian

    if isinstance(family,Gaussian):
        y = scp.stats.norm.rvs(loc=eta,scale=scale,size=n,random_state=seed)
    
    elif isinstance(family,Gamma):
        # Need to transform from mean and scale to \alpha & \beta
        # From Wood (2017), we have that
        # \phi = 1/\alpha
        # so \alpha = 1/\phi
        # From https://en.wikipedia.org/wiki/Gamma_distribution, we have that:
        # \mu = \alpha/\beta
        # \mu = 1/\phi/\beta
        # \beta = 1/\phi/\mu
        # scipy docs, say to set scale to 1/\beta.
        # see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html
        mu = family.link.fi(eta)
        alpha = 1/scale
        beta = alpha/mu  
        y = scp.stats.gamma.rvs(a=alpha,scale=(1/beta),size=n,random_state=seed)
    
    elif isinstance(family,Binomial):
        eta += binom_offset
        mu = family.link.fi(eta*scale)
        y = scp.stats.binom.rvs(1, mu, size=n,random_state=seed)

    dat = pd.DataFrame({"y":y,
                        "x0":x0,
                        "x1":x1,
                        "x2":x2,
                        "x3":x3,
                        "eta":eta})
    return dat


def sim4(n,scale,c=1,binom_offset=0,family=Gaussian(),correlate=False,seed=None):
    """
    Like ``sim3``, except that a random factor is added - second simulation performed by Wood et al., (2016).

    This is also the sixth simulation performed by gamSim() - except for the fact that c is used here to scale the contribution
    of the random factor, as was also done by Wood et al., (2016)
    
    Covariates can also be simulated to correlate with each other, following the steps outlined in supplementary materials
    E of Wood et al., (2016).

    References:

     - Gu, C. & Whaba, G., (1991). Minimizing GCV/GML scores with multiple smoothing parameters via the Newton method.
     - Wood, S. N., Pya, N., Saefken, B., (2016). Smoothing Parameter and Model Selection for General Smooth Models
     - mgcv source code: gam.sim.r

    :param scale: Standard deviation for `family='Gaussian'` else scale parameter
    :type scale: float
    :param c: Effect strength for random effect - 0 = No effect (sd=0), 1 = Maximal effect (sd=1)
    :type c: float
    :param binom_offset: Additive adjustment to log-predictor for Binomial model (5 in mgcv). Defaults to 0.
    :type binom_offset: float
    :param family: Distribution for response variable, must be: `Gaussian()`, `Gamma()`, or `Binomial()`. Defaults to `Gaussian()`
    :type family: Family, optional
    :param correlate: Whether predictor covariates should correlate or not. Defaults to False
    :type correlate: bool
    """
    np_gen = np.random.default_rng(seed)

    if correlate:
        Sigma = np.zeros((4,4)) + 0.9

        for ij in range(4):
            Sigma[ij,ij] = 1

        z = scp.stats.multivariate_normal.rvs(mean=[0 for _ in range(4)],cov=Sigma,size=n,random_state=seed)
        
        # I am a bit confused by notation in Wood et al. (2016) - they say x = cdf^{-1}(z) but I think that's not
        # what they mean, since cdf^{-1} = percent wise/quantile function which expects values between 0-1 which
        # is not the support for z. So I just use the cdf - which I think is what they mean. The resulting marginals
        # for x are uniform and all variables show the expected correlation, so it's probably correct.
        x0 = scp.stats.norm.cdf(z[:,0])
        x1 = scp.stats.norm.cdf(z[:,1])
        x2 = scp.stats.norm.cdf(z[:,2])
        x3 = scp.stats.norm.cdf(z[:,3])
    else:
        x0 = np_gen.random(n)
        x1 = np_gen.random(n)
        x2 = np_gen.random(n)
        x3 = np_gen.random(n)
    x4 = np_gen.integers(low=0,high=40,size=n)

    if c > 0:
        rind = scp.stats.norm.rvs(size=40,scale=c,random_state=seed)
    else:
        rind = np.zeros(40)

    f0 = 2* np.sin(np.pi*x0)
    f1 = np.exp(2*x1)
    f2 = 0.2*np.power(x2,11)*np.power(10*(1-x2),6)+10*np.power(10*x2,3)*np.power(1-x2,10)
    f3 = np.zeros_like(x3)
    f4 = rind[x4]

    eta = f0 + f1 + f2 + f3 + f4 # eta in truth for non-Gaussian

    if isinstance(family,Gaussian):
        y = scp.stats.norm.rvs(loc=eta,scale=scale,size=n,random_state=seed)
    
    elif isinstance(family,Gamma):
        # Need to transform from mean and scale to \alpha & \beta
        # From Wood (2017), we have that
        # \phi = 1/\alpha
        # so \alpha = 1/\phi
        # From https://en.wikipedia.org/wiki/Gamma_distribution, we have that:
        # \mu = \alpha/\beta
        # \mu = 1/\phi/\beta
        # \beta = 1/\phi/\mu
        # scipy docs, say to set scale to 1/\beta.
        # see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html
        mu = family.link.fi(eta)
        alpha = 1/scale
        beta = alpha/mu  
        y = scp.stats.gamma.rvs(a=alpha,scale=(1/beta),size=n,random_state=seed)
    
    elif isinstance(family,Binomial):
        eta += binom_offset
        mu = family.link.fi(eta*scale)
        y = scp.stats.binom.rvs(1, mu, size=n,random_state=seed)

    dat = pd.DataFrame({"y":y,
                        "x0":x0,
                        "x1":x1,
                        "x2":x2,
                        "x3":x3,
                        "x4":[f"f_{fl}" for fl in x4],
                        "eta":eta})
    return dat


def sim5(n,seed=None):
    """
    Simulates `n` data-points for a Multi-nomial model - probability of Y_i being one of K=5 classes changes smoothly as a function of variable
    x and differently so for each class - based on slightly modified versions of the original functions of Gu & Whaba (1991).

    References:
    - Gu, C. & Whaba, G., (1991). Minimizing GCV/GML scores with multiple smoothing parameters via the Newton method.
    - Wood, S. N., Pya, N., Saefken, B., (2016). Smoothing Parameter and Model Selection for General Smooth Models
    - mgcv source code: gam.sim.r
    """
    np_gen = np.random.default_rng(seed)

    x0 = np_gen.random(n)

    f0 = 2* np.sin(np.pi*x0)
    f1 = np.exp(2*x0)*0.2
    f2 = 1e-4*np.power(x0,11)*np.power(10*(1-x0),6)+10*np.power(10*x0,3)*np.power(1-x0,10)
    f3 = 1*x0 + 0.03*x0**2

    family = MULNOMLSS(4)

    mus = [np.exp(f0),np.exp(f1),np.exp(f2),np.exp(f3)]
    
    ps = np.zeros((n,5))

    for k in range(5):
        lpk = family.lp(np.zeros(n)+k, *mus)
        ps[:,k] += lpk
    
    y = np.zeros(n,dtype=int)
    
    for i in range(n):
        y[i] = int(np_gen.choice([0,1,2,3,4],p=np.exp(ps[i,:]),size=1)[0])

    dat = pd.DataFrame({"y":y,
                        "x0":x0})
    return dat


def sim6(n,family=GAUMLSS([Identity(),LOG()]),seed=None):
    """
    Simulates `n` data-points for a Gaussian or Gamma GAMMLSS model - mean and standard deviation/scale change based on 
    the original functions of Gu & Whaba (1991).

    References:
    - Gu, C. & Whaba, G., (1991). Minimizing GCV/GML scores with multiple smoothing parameters via the Newton method.
    - Wood, S. N., Pya, N., Saefken, B., (2016). Smoothing Parameter and Model Selection for General Smooth Models
    - mgcv source code: gam.sim.r

    :param family: Distribution for response variable, must be: `GAUMLSS()`, `GAMMALS()`. Defaults to `GAUMLSS([Identity(),LOG()])`
    :type family: GAMLSSFamily, optional
    """
    np_gen = np.random.default_rng(seed)

    x0 = np_gen.random(n)
    mu_sd = 2* np.sin(np.pi*x0)
    mu_mean = 0.2*np.power(x0,11)*np.power(10*(1-x0),6)+10*np.power(10*x0,3)*np.power(1-x0,10)

    mus = [mu_mean,mu_sd]

    if isinstance(family,GAUMLSS):
        y = scp.stats.norm.rvs(loc=mus[0],scale=mus[1],size=n,random_state=seed)

    elif isinstance(family,GAMMALS):
        # Need to transform from mean and scale to \alpha & \beta
        # From Wood (2017), we have that
        # \phi = 1/\alpha
        # so \alpha = 1/\phi
        # From https://en.wikipedia.org/wiki/Gamma_distribution, we have that:
        # \mu = \alpha/\beta
        # \mu = 1/\phi/\beta
        # \beta = 1/\phi/\mu
        # scipy docs, say to set scale to 1/\beta.
        # see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html

        mus[0] += 1
        mus[1] += 1

        alpha = 1/mus[1]
        beta = alpha/mus[0]  
        y = scp.stats.gamma.rvs(a=alpha,scale=(1/beta),size=n,random_state=seed)
    
    dat = pd.DataFrame({"y":y,
                        "x0":x0})
    return dat

def sim7(n,c,scale,family=Gaussian(),seed=None):
    """
    An overlap simulation with a random intercept. Two events are present that differ in their onset across
    different trials - each event triggers a response in the signal (modeled as two functions of Gu & Whaba).

    References:

     - Gu, C. & Whaba, G., (1991). Minimizing GCV/GML scores with multiple smoothing parameters via the Newton method.
     - Wood, S. N., Pya, N., Saefken, B., (2016). Smoothing Parameter and Model Selection for General Smooth Models
     - mgcv source code: gam.sim.r

    :param n: Number of trials to simulate
    :type n: int
    :param c: Effect strength for random effect - 0 = No effect (sd=0), 1 = Maximal effect (sd=1)
    :type c: float
    :param scale: Standard deviation for `family='Gaussian'` else scale parameter
    :type scale: float
    :param family: Distribution for response variable, must be: `Gaussian()`, `Gamma()`, or `Binomial()`. Defaults to `Gaussian()`
    :type family: Family, optional
    """

    np_gen = np.random.default_rng(seed)

    onsets1 = np_gen.integers(5,high=10,size=n)
    onsets2 = onsets1 + np_gen.integers(1,high=10,size=n)
    ends = onsets2 + np_gen.integers(15,high=30,size=n)
    
    if c > 0:
        rind = scp.stats.norm.rvs(size=40,scale=c,random_state=seed)
    else:
        rind = np.zeros(40)

    y = []
    time = []
    fl = []
    series = []

    for tr in range(n):
        x = np.linspace(0,1,20)
        f1 = 2* np.sin(np.pi*x)
        f2 = 0.2*np.power(x,11)*np.power(10*(1-x),6)+10*np.power(10*x,3)*np.power(1-x,10)
        f1_emb = np.zeros(ends[tr]+1)
        f2_emb = np.zeros(ends[tr]+1)
        f1_emb[0:len(f1)] += f1
        f2_emb[0:len(f2)] += f2

        # Convolve to shift event onset
        o1 = convolve_event(f1_emb,onsets1[tr])[0:len(f1_emb)]
        o2 = convolve_event(f2_emb,onsets2[tr])[0:len(f2_emb)]

        # Create & collect all variables
        t = np.arange(len(o1))*10
        y.extend(o1 + o2 + rind[tr%40])
        time.extend(t)
        fl.extend([f"l{tr%40}" for _ in range(len(o1))])
        series.extend([tr for _ in range(len(o1))])

    y = scp.stats.norm.rvs(loc=np.array(y),scale=scale,size=len(y),random_state=seed) 

    dat = pd.DataFrame({"y":y,
                        "time":time,
                        "factor":fl,
                        "series":series})
    
    return dat,onsets1,onsets2

def sim8(n,c,family=GAUMLSS([Identity(),LOG()]),seed=None):
    """
    Like sim6: Simulates `n` data-points for a Gaussian or Gamma GAMMLSS model - mean and standard deviation/scale change based on 
    the original functions of Gu & Whaba (1991). Difference is that the effect strength of the scale smoother can be manipulated by changing c!

    References:
    - Gu, C. & Whaba, G., (1991). Minimizing GCV/GML scores with multiple smoothing parameters via the Newton method.
    - Wood, S. N., Pya, N., Saefken, B., (2016). Smoothing Parameter and Model Selection for General Smooth Models
    - mgcv source code: gam.sim.r

    :param c: Effect strength for random effect - 0 = No effect (sd=0), 1 = Maximal effect (sd=1)
    :type c: float
    :param family: Distribution for response variable, must be: `GAUMLSS()`, `GAMMALS()`. Defaults to `GAUMLSS([Identity(),LOG()])`
    :type family: GAMLSSFamily, optional
    """
    np_gen = np.random.default_rng(seed)

    x0 = np_gen.random(n)
    mu_sd = 2* np.sin(np.pi*x0)
    mu_sd = 3.5 + c*mu_sd
    mu_mean = 0.2*np.power(x0,11)*np.power(10*(1-x0),6)+10*np.power(10*x0,3)*np.power(1-x0,10)

    mus = [mu_mean,mu_sd]

    if isinstance(family,GAUMLSS):
        y = scp.stats.norm.rvs(loc=mus[0],scale=mus[1],size=n,random_state=seed)

    elif isinstance(family,GAMMALS):
        # Need to transform from mean and scale to \alpha & \beta
        # From Wood (2017), we have that
        # \phi = 1/\alpha
        # so \alpha = 1/\phi
        # From https://en.wikipedia.org/wiki/Gamma_distribution, we have that:
        # \mu = \alpha/\beta
        # \mu = 1/\phi/\beta
        # \beta = 1/\phi/\mu
        # scipy docs, say to set scale to 1/\beta.
        # see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html

        mus[0] += 1
        mus[1] += 1

        alpha = 1/mus[1]
        beta = alpha/mus[0]  
        y = scp.stats.gamma.rvs(a=alpha,scale=(1/beta),size=n,random_state=seed)
    
    dat = pd.DataFrame({"y":y,
                        "x0":x0})
    return dat


def sim9(n,c=1,family=GAUMLSS([Identity(),LOG()]),seed=None):
    """
    Like ``sim4``, except for a GAMMLSS model.

    The random intercept to be selected is included in the model of the mean. I.e. the model of the mean is: f0 + f1 + c*f4 For the sd: f2 + f3.

    References:

     - Gu, C. & Whaba, G., (1991). Minimizing GCV/GML scores with multiple smoothing parameters via the Newton method.
     - Wood, S. N., Pya, N., Saefken, B., (2016). Smoothing Parameter and Model Selection for General Smooth Models
     - mgcv source code: gam.sim.r

    :param scale: Standard deviation for `family='Gaussian'` else scale parameter
    :type scale: float
    :param c: Effect strength for random effect - 0 = No effect (sd=0), 1 = Maximal effect (sd=1)
    :type c: float
    :param family: Distribution for response variable, must be: `GAUMLSS()` or `GAMMALS()`. Defaults to `GAUMLSS([Identity(),LOGb()])`
    :type family: Family, optional
    """
    np_gen = np.random.default_rng(seed)

    x0 = np_gen.random(n)
    x1 = np_gen.random(n)
    x2 = np_gen.random(n)
    x3 = np_gen.random(n)
    x4 = np_gen.integers(low=0,high=40,size=n)

    if c > 0:
        rind = scp.stats.norm.rvs(size=40,scale=c,random_state=seed)
    else:
        rind = np.zeros(40)

    f0 = 2* np.sin(np.pi*x0)
    f1 = np.exp(2*x1)
    f2 = 0.2*np.power(x2,11)*np.power(10*(1-x2),6)+10*np.power(10*x2,3)*np.power(1-x2,10)
    f3 = np.zeros_like(x3)
    f4 = rind[x4]

    mu_mean = f0 + f1 + f4
    mu_sd = 5 + f2 + f3

    mus = [mu_mean,mu_sd]

    if isinstance(family,GAUMLSS):
        y = scp.stats.norm.rvs(loc=mus[0],scale=mus[1],size=n,random_state=seed)
    
    elif isinstance(family,GAMMALS):
        # Need to transform from mean and scale to \alpha & \beta
        # From Wood (2017), we have that
        # \phi = 1/\alpha
        # so \alpha = 1/\phi
        # From https://en.wikipedia.org/wiki/Gamma_distribution, we have that:
        # \mu = \alpha/\beta
        # \mu = 1/\phi/\beta
        # \beta = 1/\phi/\mu
        # scipy docs, say to set scale to 1/\beta.
        # see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html

        mus[0] += 1

        alpha = 1/mus[1]
        beta = alpha/mus[0]  
        y = scp.stats.gamma.rvs(a=alpha,scale=(1/beta),size=n,random_state=seed)

    dat = pd.DataFrame({"y":y,
                        "x0":x0,
                        "x1":x1,
                        "x2":x2,
                        "x3":x3,
                        "x4":[f"f_{fl}" for fl in x4]})
    return dat


def sim10(n,c=1,family=GAUMLSS([Identity(),LOG()]),seed=None):
    """
    Like ``sim9``, except that c is used to scale effect of f0.

    I.e. the model of the mean is: c*f0 + f1 + f4 For the sd: f2 + f3.

    References:

     - Gu, C. & Whaba, G., (1991). Minimizing GCV/GML scores with multiple smoothing parameters via the Newton method.
     - Wood, S. N., Pya, N., Saefken, B., (2016). Smoothing Parameter and Model Selection for General Smooth Models
     - mgcv source code: gam.sim.r

    :param scale: Standard deviation for `family='Gaussian'` else scale parameter
    :type scale: float
    :param c: Effect strength for random effect - 0 = No effect (sd=0), 1 = Maximal effect (sd=1)
    :type c: float
    :param family: Distribution for response variable, must be: `GAUMLSS()` or `GAMMALS()`. Defaults to `GAUMLSS([Identity(),LOG()])`
    :type family: Family, optional
    """
    np_gen = np.random.default_rng(seed)

    x0 = np_gen.random(n)
    x1 = np_gen.random(n)
    x2 = np_gen.random(n)
    x3 = np_gen.random(n)
    x4 = np_gen.integers(low=0,high=40,size=n)

    rind = scp.stats.norm.rvs(size=40,scale=1,random_state=seed)

    f0 = 2* np.sin(np.pi*x0)
    f1 = np.exp(2*x1)
    f2 = 0.2*np.power(x2,11)*np.power(10*(1-x2),6)+10*np.power(10*x2,3)*np.power(1-x2,10)
    f3 = np.zeros_like(x3)
    f4 = rind[x4]

    mu_mean = c*f0 + f1 + f4
    mu_sd = 5 + f2 + f3

    mus = [mu_mean,mu_sd]

    if isinstance(family,GAUMLSS):
        y = scp.stats.norm.rvs(loc=mus[0],scale=mus[1],size=n,random_state=seed)
    
    elif isinstance(family,GAMMALS):
        # Need to transform from mean and scale to \alpha & \beta
        # From Wood (2017), we have that
        # \phi = 1/\alpha
        # so \alpha = 1/\phi
        # From https://en.wikipedia.org/wiki/Gamma_distribution, we have that:
        # \mu = \alpha/\beta
        # \mu = 1/\phi/\beta
        # \beta = 1/\phi/\mu
        # scipy docs, say to set scale to 1/\beta.
        # see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html

        mus[0] += 1

        alpha = 1/mus[1]
        beta = alpha/mus[0]  
        y = scp.stats.gamma.rvs(a=alpha,scale=(1/beta),size=n,random_state=seed)

    dat = pd.DataFrame({"y":y,
                        "x0":x0,
                        "x1":x1,
                        "x2":x2,
                        "x3":x3,
                        "x4":[f"f_{fl}" for fl in x4]})
    return dat


def sim11(n,scale,c=1,binom_offset=0,n_ranef=40,family=Gaussian(),seed=None):
    """
    Like ``sim4``, except that a random smooth of variable `x0` is added - extension of the second simulation performed by Wood et al., (2016).

    c is used here to scale the contribution of the random smooth. Setting it to 0 means the ground truth is maximally wiggly. Setting it to 1
    means the random smooth is actually a random intercept.

    References:

     - Gu, C. & Whaba, G., (1991). Minimizing GCV/GML scores with multiple smoothing parameters via the Newton method.
     - Wood, S. N., Pya, N., Saefken, B., (2016). Smoothing Parameter and Model Selection for General Smooth Models
     - mgcv source code: gam.sim.r

    :param scale: Standard deviation for `family='Gaussian'` else scale parameter
    :type scale: float
    :param c: Effect strength for random smooth - 0 = Maximally wiggly, 1 = ground truth is random intercept
    :type c: float
    :param binom_offset: Additive adjustment to log-predictor for Binomial model (5 in mgcv). Defaults to 0.
    :type binom_offset: float
    :param n_ranef: Number of levels for the random smooth term. Defaults to 40.
    :type n_ranef: int
    :param family: Distribution for response variable, must be: `Gaussian()`, `Gamma()`, or `Binomial()`. Defaults to `Gaussian()`
    :type family: Family, optional
    """
    np_gen = np.random.default_rng(seed)

    x0 = np_gen.random(n)
    x1 = np_gen.random(n)
    x2 = np_gen.random(n)
    x3 = np_gen.random(n)
    x4 = np_gen.integers(low=0,high=n_ranef-1,size=n)

    f0 = 2* np.sin(np.pi*x0)
    f1 = np.exp(2*x1)
    f2 = 0.2*np.power(x2,11)*np.power(10*(1-x2),6)+10*np.power(10*x2,3)*np.power(1-x2,10)
    f3 = np.zeros_like(x3)
    f4 = np.zeros_like(x0)

    c = max(1e-7,c)
    
    # Set up random smooth sampler for smooth of x0
    # Based on prior assumption discussed by Wood (2017)
    fs_dat = pd.DataFrame({"x0":x0,
                        "y":np_gen.random(n)})

    fs_formula = Formula(lhs=lhs("y"),
                            terms=[f(["x0"],identifiable=False,nk=10,penalty=[PenType.DIFFERENCE],pen_kwargs=[{"m":1}])],
                            data=fs_dat)

    fs_model = GAMM(fs_formula,Gaussian())
    fs_model.formula.build_penalties()

    mmat = fs_model.get_mmat()
    cov = fs_formula.cov_flat
    S = fs_formula.penalties[0].S_J

    C, Srp, Drp, IRrp, rms1, rms2, rp_rank = reparam(mmat,S,cov,identity=True,scale=False,QR=True)
    
    mmat_RP = mmat @ C
    mmat_RP[:,[-1]] = 1
    
    V = (Srp/(c*1e7)).toarray()
    
    V[-1,-1] = 1
    #print(V)

    for l4 in np.unique(x4):
        if not seed is None:
            sample = scp.stats.multivariate_normal.rvs(mean=[0 for _ in range(Srp.shape[1])],cov=V,size=1,random_state=seed+l4)
        else:
            sample = scp.stats.multivariate_normal.rvs(mean=[0 for _ in range(Srp.shape[1])],cov=V,size=1,random_state=None)
        #print(sample)
        if c >= 1: # Random smooth is random intercept
            sample[:-1] = 0

        fl4 = mmat_RP@sample
        f4[x4 == l4] = fl4[x4 == l4]


    eta = f0 + f1 + f2 + f3 + f4 # eta in truth for non-Gaussian

    if isinstance(family,Gaussian):
        y = scp.stats.norm.rvs(loc=eta,scale=scale,size=n,random_state=seed)
    
    elif isinstance(family,Gamma):
        # Need to transform from mean and scale to \alpha & \beta
        # From Wood (2017), we have that
        # \phi = 1/\alpha
        # so \alpha = 1/\phi
        # From https://en.wikipedia.org/wiki/Gamma_distribution, we have that:
        # \mu = \alpha/\beta
        # \mu = 1/\phi/\beta
        # \beta = 1/\phi/\mu
        # scipy docs, say to set scale to 1/\beta.
        # see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html
        mu = family.link.fi(eta)
        alpha = 1/scale
        beta = alpha/mu  
        y = scp.stats.gamma.rvs(a=alpha,scale=(1/beta),size=n,random_state=seed)
    
    elif isinstance(family,Binomial):
        eta += binom_offset
        mu = family.link.fi(eta*scale)
        y = scp.stats.binom.rvs(1, mu, size=n,random_state=seed)

    dat = pd.DataFrame({"y":y,
                        "x0":x0,
                        "x1":x1,
                        "x2":x2,
                        "x3":x3,
                        "x4":[f"f_{fl}" for fl in x4],
                        "eta":eta})
    return dat

def sim12(n,c=1,n_ranef=40,family=GAUMLSS([Identity(),LOG]),seed=None):
    """
    Like ``sim11``, except for GAMLSS models. ``x0``, ``x1``, and ``x4`` impact the mean - the remaining variables impact the scale parameter.

    c is used here to scale the contribution of the random smooth (which is included in the model fo the mean). Setting it to 0 means the ground truth is maximally wiggly. Setting it to 1
    means the random smooth is actually a random intercept.

    References:

     - Gu, C. & Whaba, G., (1991). Minimizing GCV/GML scores with multiple smoothing parameters via the Newton method.
     - Wood, S. N., Pya, N., Saefken, B., (2016). Smoothing Parameter and Model Selection for General Smooth Models
     - mgcv source code: gam.sim.r

    :param c: Effect strength for random smooth - 0 = Maximally wiggly, 1 = ground truth is random intercept
    :type c: float
    :param n_ranef: Number of levels for the random smooth term. Defaults to 40.
    :type n_ranef: int
    :param family: Distribution for response variable, must be: `GAUMLSS()` or `GAMMALS()`. Defaults to `GAUMLSS([Identity(),LOGb()])`
    :type family: Family, optional
    """
    np_gen = np.random.default_rng(seed)

    x0 = np_gen.random(n)
    x1 = np_gen.random(n)
    x2 = np_gen.random(n)
    x3 = np_gen.random(n)
    x4 = np_gen.integers(low=0,high=n_ranef-1,size=n)

    f0 = 2* np.sin(np.pi*x0)
    f1 = np.exp(2*x1)
    f2 = 0.2*np.power(x2,11)*np.power(10*(1-x2),6)+10*np.power(10*x2,3)*np.power(1-x2,10)
    f3 = np.zeros_like(x3)
    f4 = np.zeros_like(x0)

    c = max(1e-7,c)
    
    # Set up random smooth sampler for smooth of x0
    # Based on prior assumption discussed by Wood (2017)
    fs_dat = pd.DataFrame({"x0":x0,
                        "y":np_gen.random(n)})

    fs_formula = Formula(lhs=lhs("y"),
                            terms=[f(["x0"],identifiable=False,nk=10,penalty=[PenType.DIFFERENCE],pen_kwargs=[{"m":1}])],
                            data=fs_dat)

    fs_model = GAMM(fs_formula,Gaussian())
    fs_model.formula.build_penalties()

    mmat = fs_model.get_mmat()
    cov = fs_formula.cov_flat
    S = fs_formula.penalties[0].S_J

    C, Srp, Drp, IRrp, rms1, rms2, rp_rank = reparam(mmat,S,cov,identity=True,scale=False,QR=True)
    
    mmat_RP = mmat @ C
    mmat_RP[:,[-1]] = 1
    
    V = (Srp/(c*1e7)).toarray()
    
    V[-1,-1] = 1
    #print(V)

    for l4 in np.unique(x4):
        if not seed is None:
            sample = scp.stats.multivariate_normal.rvs(mean=[0 for _ in range(Srp.shape[1])],cov=V,size=1,random_state=seed+l4)
        else:
            sample = scp.stats.multivariate_normal.rvs(mean=[0 for _ in range(Srp.shape[1])],cov=V,size=1,random_state=None)
        #print(sample)
        if c >= 1: # Random smooth is random intercept
            sample[:-1] = 0

        fl4 = mmat_RP@sample
        f4[x4 == l4] = fl4[x4 == l4]

    eta_mean = f0 + f1 + f4
    eta_sd = f2 + f3
    
    eta_sd *= 0.1

    if isinstance(family.links[0],LOG):
        eta_mean *= 0.5

    mus = [family.links[0].fi(eta_mean),family.links[1].fi(eta_sd)]
    #print(mus[1])

    if isinstance(family,GAUMLSS):
        y = scp.stats.norm.rvs(loc=mus[0],scale=mus[1],size=n,random_state=seed)
    
    elif isinstance(family,GAMMALS):
        # Need to transform from mean and scale to \alpha & \beta
        # From Wood (2017), we have that
        # \phi = 1/\alpha
        # so \alpha = 1/\phi
        # From https://en.wikipedia.org/wiki/Gamma_distribution, we have that:
        # \mu = \alpha/\beta
        # \mu = 1/\phi/\beta
        # \beta = 1/\phi/\mu
        # scipy docs, say to set scale to 1/\beta.
        # see: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gamma.html
        
        alpha = 1/mus[1]
        beta = alpha/mus[0]  
        y = scp.stats.gamma.rvs(a=alpha,scale=(1/beta),size=n,random_state=seed)

    dat = pd.DataFrame({"y":y,
                        "x0":x0,
                        "x1":x1,
                        "x2":x2,
                        "x3":x3,
                        "x4":[f"f_{fl}" for fl in x4],
                        "eta_mean":eta_mean,
                        "eta_scale":eta_sd})
    return dat


import copy
import math
from .sim import np,pd,scp
from mssm.models import *

################################## Contains functions to extract useful information from GAMM & GAMMLSS models ##################################


def eval_coverage(model:GAMM or GAMMLSS,pred_dat,dist_par=0,target:float or [float]=0.0,use:[int]=None,alpha=0.05,whole_function=False,n_ps=10000,seed=None):
    """Evaluate CI coverage of ``target`` function over domain defined by ``pred_dat``.

    :param model: ``GAMM`` or ``GAMMLSS`` model.
    :type model: GAMM or GAMMLSS
    :param pred_dat: ``pandas.DataFrame`` with data used to compute model predictions to be compared against target as well as CI.
    :type pred_dat: pd.Dataframe
    :param dist_par: The index corresponding to the parameter for which to make the prediction (e.g., 0 = mean) - only necessary if a GAMMLSS model is provided, defaults to 0
    :type dist_par: int, optional
    :param target: Target function. Can either be set to a float (e.g., 0.0 if the target function is believed to be zero everywhere across the domain) or a ``list``/``np.array ``of floats. In the latter case the shape of ``target`` must be a flattened 1D array. defaults to 0.0
    :type target: float or [float], optional
    :param use: The indices corresponding to the terms that should be used to obtain the prediction or ``None``
    in which case all terms will be used.
    :type use: list[int] or None
    :param alpha: The alpha level to use for the standard error calculation. Specifically, 1 - (`alpha`/2) will be used to determine the critical cut-off value according to a N(0,1).
    :type alpha: float, optional
    :param whole_function: Whether or not to adjuste the point-wise CI to behave like whole-function (based on Wood, 2017; section 6.10.2 and Simpson, 2016). Defaults to False.
    :type whole_function: bool, optional
    :param n_ps: How many samples to draw from the posterior in case the point-wise CI is adjusted to behave like a whole-function CI.
    :type n_ps: int, optional
    :param seed: Can be used to provide a seed for the posterior sampling step in case the point-wise CI is adjusted to behave like a whole-function CI.
    :type seed: int or None, optional
    :return: A tuple with three elements. First is a bool, indicating whether the target function is covered by the CI at every evaluated value of the domain. Second is the (average) coverage across the entire domain.
    Third is a boolean array indicating for every evaluated value whether the corresponding value of target falls within the CI boundaries.
    :rtype: (bool,float,[bool])
    """

    # Compute model prediction and CI boundaries
    if isinstance(model,GAMMLSS):
        pred,_,b = model.predict(dist_par,use,pred_dat,ci=True,whole_interval=whole_function,alpha=alpha,n_ps=n_ps,seed=seed)
    else:
        pred,_,b = model.predict(use,pred_dat,ci=True,whole_interval=whole_function,alpha=alpha,n_ps=n_ps,seed=seed)

    # Compute upper and lower boundaries
    UB = pred + b
    LB = pred - b

    # Check for how many evaluated values CI contains ground truth.
    IN_CI = (UB >= target) & (LB <= target)

    full_coverage = True # Function covered fully?
    coverage = 1

    # Function is not covered at every value of domain by CI boundaries, re-compute coverage across domain
    if False in np.unique(IN_CI):
        full_coverage = False
        coverage = np.sum(IN_CI)/len(IN_CI)
    
    return full_coverage,coverage,IN_CI


def get_term_coef(model:GAMM or GAMMLSS,which:[int],dist_par=0):
    """Get the coefficients associated with a specific term included in the ``Formula`` of ``model``. Useful to extract for example
    the estimated random intercepts from a random effect model.

    Note, coefficients will be in order of the model matrix columns. For a random intercept term (``ri()``) this implies that the first coefficient
    in the returned vector will correspond to the first encoded level of the random factor. Usually, ``mssm`` determines the level-ordering automatically and
    the actual factor-level corresponding to the first encoded level can be determined by inspecting the code-book returned from ``formula.get_factor_codings()``.


    :param model: ``GAMM`` or ``GAMMLSS`` model.
    :type model: GAMM or GAMMLSS
    :param which: Index corresponding to the term in the model's formula for which the coefficients should be extracted.
    :type which: [int]
    :param dist_par: The index corresponding to the parameter for which to make the prediction (e.g., 0 = mean) - only necessary if a GAMMLSS model is provided, defaults to 0
    :type dist_par: int, optional
    """

    # Get model-matrix that was used for fitting.
    model_mat = model.get_mmat(use_terms=which)

    if isinstance(model,GAMMLSS):
        model_mat = model_mat[dist_par]
    
    # Find coefficient indices corresponding to indicated terms:
    coef_idx = model_mat.sum(axis=0) != 0

    # Return corresponding coefficients.
    if isinstance(model,GAMMLSS):
        split_coef = np.split(model.coef,model.coef_split_idx)
        return split_coef[dist_par][coef_idx]
    else:
        return model.coef[coef_idx]
    




import matplotlib.pyplot as plt
import matplotlib.cm as cmx
import matplotlib
from matplotlib import colors
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
import copy
import math
from .sim import np,pd,scp
from mssm.models import *

################################## Contains functions to visualize and validate GAMM & GAMMLSS models ##################################


def __get_data_limit_counts(formula,pred_dat,cvars,by):
    """Checks for every row in the data used for prediction, whether continuous variables are within training data limits.
     
    Also finds how often each combination of continuous variables exists in trainings data.

    :param formula: A GAMM Formula, model must have been fit.
    :type formula: Formula
    :param pred_dat: pandas DataFrame holding prediction data.
    :type pred_dat: pandas.Dataframe
    :param cvars: A list of the continuous variables to take into account.
    :type cvars: [str]
    :param by: A list of categorical variables associated with a smooth term, i.e., if a smooth has a different shape for different levels of a factor or a prediction.
    :type by: [str]
    :return: Three vectors + list. First contains bool indicating whether all continuous variables in prediction data row had values within training limits. Second contains all
    unique combinations of continuous variable values in training set. Third contains count for each unique combination in training set. Final list holds names of continuous variables in the order of the columns of the second vector.
    :rtype: tuple
    """
    
    _,pred_cov,_,_,_,_,_ = formula.encode_data(pred_dat,prediction=True)

    # Find continuous predictors and categorical ones
    pred_cols = pred_dat.columns
    cont_idx = [formula.get_var_map()[var] for var in pred_cols if formula.get_var_types()[var] == VarType.NUMERIC and var in cvars]
    cont_vars = [var for var in pred_cols if formula.get_var_types()[var] == VarType.NUMERIC and var in cvars]
    factor_idx = []

    if not by is None:
        factor_idx = [formula.get_var_map()[var] for var in pred_cols if formula.get_var_types()[var] == VarType.FACTOR and var in by]

    # Get sorted encoded cov structure for prediction and training data containing continuous variables
    sort_pred = pred_cov[:,cont_idx]
    sort_train = formula.cov_flat[:,cont_idx]

    if len(factor_idx) > 0 and not by is None:
        # Now get all columns corresponding to factor variables so that we can check
        # which rows in the trainings data belong to conditions present in the pred data.
        sort_cond_pred = pred_cov[:,factor_idx]
        sort_cond_train = formula.cov_flat[:,factor_idx]

        # Now get unique encoded rows - only considering factor variables
        pred_cond_unique = np.unique(sort_cond_pred,axis=0)

        train_cond_unq,train_cond_inv = np.unique(sort_cond_train,axis=0,return_inverse=True)

        # Check which training conditions are present in prediction
        train_cond_unq_exists = np.array([(train == pred_cond_unique).all(axis=1).any() for train in train_cond_unq])

        # Now get part of training cov matching conditions in prediction data
        sort_train = sort_train[train_cond_unq_exists[train_cond_inv],:]

    # Check for each combination in continuous prediction columns whether the values are within
    # min and max of the respective trainings columns
    pred_in_limits = np.ones(len(sort_pred),dtype=bool)
    sort_pred_scaled = copy.deepcopy(sort_pred)
    sort_train_scaled = copy.deepcopy(sort_train)
    
    # Based on exclude.too.far function in mgcv - scale new & train data to be in unit square
    # see: https://rdrr.io/cran/mgcv/src/R/plots.r#sym-exclude.too.far
    for cont_i in range(sort_pred.shape[1]):
        
        min_pred = min(sort_pred_scaled[:,cont_i])
        sort_pred_scaled[:,cont_i] -= min_pred
        sort_train_scaled[:,cont_i] -= min_pred
        max_pred = max(sort_pred_scaled[:,cont_i])
        sort_pred_scaled[:,cont_i] /= max_pred
        sort_train_scaled[:,cont_i] /= max_pred

    # Then find minimum distance to any data point for each prediction value
    dist = np.array([min(np.linalg.norm(sort_train_scaled - sort_pred_scaled[predi,:],axis=1)) for predi in range(sort_pred.shape[0])])
    pred_in_limits[dist > 0.1] = False

    # Now find the counts in the training data for each combination of continuous variables
    train_unq,train_unq_counts = np.unique(sort_train,axis=0,return_counts=True)
    
    return pred_in_limits,train_unq,train_unq_counts.astype(float),cont_vars


def __pred_plot(pred,b,tvars,pred_in_limits,x1,x2,x1_exp,ci,n_vals,ax,_cmp,col,ylim,link,legend_label):
    """Internal function to visualize a univariate smooth of covariate `x1` or a tensor smooth of covariate `x1` and `x2`.

    Called by :func:`plot`, :func:`plot_fitted`, and :func:`plot_diff`.

    :param pred: Vector holding model prediction
    :type pred: [float]
    :param b: Vector holding standard error that needs to be added to/subtracted from `pred` to obtain ci boundaries.
    :type b: [float]
    :param tvars: List of variables to be visualized - contains one string for univariate smooths, two for tensor smooths.
    :type tvars: [str]
    :param pred_in_limits: bolean vector indicating which prediction was obtained for a covariate combination within data limits.
    :type pred_in_limits: [bool]
    :param x1: Unique values of covariate x1
    :type x1: [float]
    :param x2: Unique values of covariate x2
    :type x2: [float]
    :param x1_exp: For univariate smooth like `x1`, for tensor smooth this holds `x1` for each value of `x2`
    :type x1_exp: [float]
    :param ci: Same as `x1_exp` for `x2`
    :type ci: [float]
    :param n_vals: Number of values use to create each marginal covariate
    :type n_vals: int
    :param ax: matplotlib.axis to plot on
    :type ax: matplotlib.axis
    :param _cmp: matplotlib colormap
    :type _cmp: matplotlib.colormap
    :param col: color to use for univariate plot, float in [0,1]
    :type col: float
    :param ylim: limits for y-axis
    :type ylim: (float,float)
    :param link: Link function of model.
    :type link: Link
    :param legend_label: Legend label to pass to univariate smooths.
    :type legend_label: str
    """
    
    # Handle tensor smooth case
    if len(tvars) == 2:
        T_pred = pred.reshape(n_vals,n_vals)

        if not link is None:
            T_pred = link.fi(T_pred)

        # Mask anything out of data limits.
        if not pred_in_limits is None:
            T_pred = np.ma.array(T_pred, mask = (pred_in_limits == False))

        T_pred = T_pred.T

        halfrange = None
        if not ylim is None:
            halfrange = np.max(np.abs(ylim))
        if ci:
            # Mask anything where CI contains zero
            f1 = ax.contourf(x1,x2,T_pred,levels=n_vals,cmap=_cmp,norm=colors.CenteredNorm(halfrange=halfrange),alpha=0.4)
            T_pred = np.ma.array(T_pred, mask= ((pred + b) > 0) & ((pred - b) < 0))
        
        # Plot everything (outside ci)
        ff = ax.contourf(x1,x2,T_pred,levels=n_vals,cmap=_cmp,norm=colors.CenteredNorm(halfrange=halfrange))
        ll = ax.contour(x1,x2,T_pred,colors="grey")

    elif len(tvars) == 1:

        # Form prediciton + CIs
        x = x1_exp
        y = pred
        if ci:
            cu = pred + b
            cl = pred - b

        # transformation applied to ci boundaries - NOT b!
        if not link is None:
            y = link.fi(y)
            if ci:
                cu = link.fi(cu)
                cl = link.fi(cl)

        # Hide everything outside data limits
        if not pred_in_limits is None:
            x = x[pred_in_limits]
            y = y[pred_in_limits]
            if ci:
                cu = cu[pred_in_limits]
                cl = cl[pred_in_limits]

        if ci:
            ax.fill([*x,*np.flip(x)],
                    [*(cu),*np.flip(cl)],
                    color=_cmp(col),alpha=0.5)
            
        ax.plot(x,y,color=_cmp(col),label=legend_label)

        if not ylim is None:
            ax.set_ylim(ylim)

def plot(model:GAMM or GAMMLSS,which:[int] or None = None, dist_par=0, n_vals:int = 30,ci=None,
         ci_alpha=0.05,use_inter=False,whole_interval=False,n_ps=10000,seed=None,cmp:str or None = None,
         plot_exist=True,plot_exist_style='both',response_scale=False,axs=None,
         fig_size=(6/2.54,6/2.54),math_font_size = 9,math_font = 'cm',
         ylim=None,prov_cols=None):
    """Helper function to plot all smooth functions estimated by a `GAMM` or `GAMMLSS` model.

    Smooth functions are automatically evaluated over a range of `n_values` spaced equally to cover their entire covariate.
    For tensor smooths a `n_values`*`n_values` grid is created. Visualizations can be obtained on the scale of the linear predictor (the default), but also
    on what is often referred to as the 'response'-scale - corresponding to the estimated mean of the RVs modeled by the model.

    To simply obtain visualizations of all smooth terms estimated, it is sufficient to call::

        plot(model) # or plot(model,dist_par=0) in case of a GAMMLSS model
    
    This will visualize all smooth terms estimated by the model and automatically determine whether confidence intervals should be drawn or not (by default CIs
    are only visualized for fixed effects). Note that, for tensor smooths, areas of the smooth for which the CI contains zero will be visualized with low opacity
    if the CI is to be visualized.

    References:

    - Wood, S. N. (2017). Generalized Additive Models: An Introduction with R, Second Edition (2nd ed.).
    - Simpson, G. (2016). Simultaneous intervals for smooths revisited.

    :param model: The estimated GAMM or GAMMLSS model for which the visualizations are to be obtained
    :type model: GAMM or GAMMLSS
    :param which: The indices corresponding to the smooth that should be visualized or ``None`` in which case all smooth terms will be visualized, defaults to None
    :type which: [int] or None, optional
    :param dist_par: The index corresponding to the parameter for which to make the prediction (e.g., 0 = mean) - only necessary if a GAMMLSS model is provided, defaults to 0
    :type dist_par: int, optional
    :param n_vals: Number of covariate values over which to evaluate the function. Will be **2 for tensor smooths, defaults to 30
    :type n_vals: int, optional
    :param ci: Whether the standard error ``se`` for credible interval (CI; see  Wood, 2017) calculation should be computed and used to visualize CIs. The CI is then [``pred`` - ``se``, ``pred`` + ``se``], defaults to None
    in which case the CI will be visualized for fixed effects but not for random smooths
    :type ci: bool or None, optional
    :param ci_alpha: The alpha level to use for the standard error calculation. Specifically, 1 - (``alpha``/2) will be used to determine the critical cut-off value according to a N(0,1), defaults to 0.05
    :type ci_alpha: float, optional
    :param use_inter: Whether or not the standard error for CIs should be computed based on just the smooth or based on the smooth + the model intercept - the latter results in better coverage for strongly penalized functions (see Wood, 2017), defaults to False
    :type use_inter: bool, optional
    :param whole_interval: Whether or not to adjuste the point-wise CI to behave like whole-interval (based on Wood, 2017; section 6.10.2 and Simpson, 2016), defaults to False
    :type whole_interval: bool, optional
    :param n_ps: How many samples to draw from the posterior in case the point-wise CI is adjusted to behave like a whole-interval CI, defaults to 10000
    :type n_ps: int, optional
    :param seed: Can be used to provide a seed for the posterior sampling step in case the point-wise CI is adjusted to behave like a whole-interval CI, defaults to None
    :type seed: int, optional
    :param cmp: string corresponding to name for a matplotlib colormap, defaults to None in which case it will be set to 'RdYlBu_r'.
    :type cmp: str or None, optional
    :param plot_exist: Whether or not an indication of the data distribution should be provided. For univariate smooths setting this to True will add a rug-plot to the
    bottom, indicating for which covariate values samples existed in the training data. For tensor smooths setting this to true will result in a 2d scatter rug plot being added and/or values outside of data limits being hidden, defaults to True
    :type plot_exist: bool, optional
    :param plot_exist_style: Determines the style of the data distribution indication for smooths. Must be 'rug', 'hide',or 'both'. 'both' will both add the rug-plot and hide values out of data limits, defaults to 'both'
    :type plot_exist_style: str, optional
    :param response_scale: Whether or not predictions and CIs should be shown on the scale of the model predictions (linear scale) or on the 'response-scale' i.e., the scale of the mean, defaults to False
    :type response_scale: bool, optional
    :param axs: A list of matplotlib.axis on which Figures should be drawn, defaults to None in which case axis will be created by the function and plot.show() will be called at the end
    :type axs: [matplotlib.axis], optional
    :param fig_size: Tuple holding figure size, which will be used to determine the size of the figures created if `axs=None`, defaults to (6/2.54,6/2.54)
    :type fig_size: tuple, optional
    :param math_font_size: Font size for math notation, defaults to 9
    :type math_font_size: int, optional
    :param math_font: Math font to use, defaults to 'cm'
    :type math_font: str, optional
    :param ylim: Tuple holding y-limits, defaults to None in which case y_limits will be inferred from the predictions made
    :type ylim: (float,float), optional
    :param prov_cols: A float or a list (in case of a smooth with a `by` argument) of floats in [0,1]. Used to get a color for unicariate smooth terms, defaults to None in which case colors will be selected automatically depending on whether the smooth has a `by` keyword or not
    :type prov_cols: float or [float], optional
    :raises ValueError: If fewer matplotlib axis are provided than the number of figures that would be created
    """

    if isinstance(model,GAMMLSS):
        # Set up everything for that we can plot all smooth terms for distribution parameter `dist_par`.
        model.formula = model.formulas[dist_par]
    
    # Get all necessary information from the model formula
    terms = model.formula.get_terms()
    stidx = model.formula.get_smooth_term_idx()

    varmap = model.formula.get_var_map()
    vartypes = model.formula.get_var_types()
    varmins = model.formula.get_var_mins()
    varmaxs = model.formula.get_var_maxs()
    code_factors = model.formula.get_coding_factors()
    factor_codes = model.formula.get_factor_codings()

    # Default colormap
    if cmp is None:
        cmp = 'RdYlBu_r'
        
    _cmp = matplotlib.colormaps[cmp]

    if not which is None:
        stidx = which

    # Check number of figures matches axis
    n_figures = 0
    for sti in stidx:
        if isinstance(terms[sti],fs):
            n_figures +=1
        else:
            if not terms[sti].by is None:
                n_figures += len(code_factors[terms[sti].by])

            else:
                n_figures += 1
    
    if not axs is None and len(axs) != n_figures:
        raise ValueError(f"{n_figures} plots would be created, but only {len(axs)} axes were provided!")

    # if nothing is provided, create figures + axis
    figs = None
    if axs is None:
        figs = [plt.figure(figsize=fig_size,layout='constrained') for _ in range(n_figures)]
        axs = [fig.add_subplot(1,1,1) for fig in figs]
    
    axi = 0

    for sti in stidx:

        # Start by generating prediction data for the current smooth term.
        tvars = terms[sti].variables
        pred_dat = {}
        x1_exp = []
        if len(tvars) == 2:
            # Set up a grid of n_vals*n_vals
            x1 = np.linspace(varmins[tvars[0]],varmaxs[tvars[0]],n_vals)
            x2 = np.linspace(varmins[tvars[1]],varmaxs[tvars[1]],n_vals)

            x2_exp = []

            for x1v in x1:
                for x2v in x2:
                    x1_exp.append(x1v)
                    x2_exp.append(x2v)
            
            pred_dat[tvars[0]] = x1_exp
            pred_dat[tvars[1]] = x2_exp
        
        elif len(tvars) == 1:
            x1 = None
            x2 = None
            # Simply set up x1_exp directly.
            x1_exp = np.linspace(varmins[tvars[0]],varmaxs[tvars[0]],n_vals)
            pred_dat[tvars[0]] = x1_exp
        else:
            continue
        
        # Now fill the data used for prediction with placeholders for all other variables
        # included in the model. These will be ignored for the prediction.
        if terms[sti].by is None and terms[sti].binary is None:
            for vari in varmap.keys():
                if vari in terms[sti].variables:
                    continue
                else:
                    if vartypes[vari] == VarType.FACTOR:
                        if vari in model.formula.get_subgroup_variables():
                            pred_dat[vari.split(":")[0]] = [code_factors[vari][0] for _ in range(len(x1_exp))]
                        else:
                            pred_dat[vari] = [code_factors[vari][0] for _ in range(len(x1_exp))]
                    else:
                        pred_dat[vari] = [0 for _ in range(len(x1_exp))]
            
            pred_dat_pd = pd.DataFrame(pred_dat)
            
            use_ci = ci
            if use_ci is None:
                use_ci = True

            # Add intercept for prediction - remember to subtract it later
            use = [sti]
            if use_inter:
                use = [0,sti]

            if isinstance(model,GAMMLSS):
                # Reset formula to prevent any problems with the call to predict, since the GAMMLSS class might
                # change this attribute itself.
                model.formula = None
                
                pred,_,b= model.predict(dist_par,use,pred_dat_pd,ci=use_ci,alpha=ci_alpha,whole_interval=whole_interval,n_ps=n_ps,seed=seed)

                # Set formula again.
                model.formula = model.formulas[dist_par]
            else:
                pred,_,b= model.predict(use,pred_dat_pd,ci=use_ci,alpha=ci_alpha,whole_interval=whole_interval,n_ps=n_ps,seed=seed)

            # Subtract intercept from prediction - it was just used to adjust se
            if use_inter:
                _cf,_ = model.get_pars()
                pred -= _cf[0]

            # Compute data limits and anything needed for rug plot
            plot_in_limits = None
            if plot_exist:
                pred_in_limits,train_unq,train_unq_counts,cont_vars = __get_data_limit_counts(model.formula,pred_dat_pd,tvars,None)

            if plot_exist and (plot_exist_style == "both" or plot_exist_style == "hide"):
                plot_in_limits = pred_in_limits
            
            # Prepare link to transform prediction + ci to response-scale
            link = None
            if response_scale:
                if isinstance(model,GAMMLSS):
                    link = model.family.links[dist_par]
                else:
                    link = model.family.link

            # Now plot
            __pred_plot(pred,b,tvars,plot_in_limits,x1,x2,x1_exp,use_ci,n_vals,axs[axi],_cmp,0.7 if prov_cols is None else prov_cols,ylim,link,None)

            # Specify labels and add rug plots if requested
            if len(tvars) == 1:
                axs[axi].set_ylabel('$f(' + tvars[0] + ')$',math_fontfamily=math_font,size=math_font_size,fontweight='bold')
                axs[axi].set_xlabel(tvars[0],fontweight='bold')
                axs[axi].spines['top'].set_visible(False)
                axs[axi].spines['right'].set_visible(False)

                if plot_exist:
                    
                    #train_unq_counts[train_unq_counts > 0] = 1 
                    #pred_range = np.abs(np.max(pred) - np.min(pred))*0.025
                    x_counts = np.ndarray.flatten(train_unq[:,[cvar ==tvars[0] for cvar in cont_vars]])
                    #x_range = np.abs(np.max(x_counts) - np.min(x_counts))
                    
                    axs[axi].scatter(x_counts,[axs[axi].get_ylim()[0]]*len(x_counts),marker='|',color='black',linewidths=0.25)
            
            elif len(tvars) == 2:
                axs[axi].set_ylabel(tvars[1],fontweight='bold')
                axs[axi].set_xlabel(tvars[0],fontweight='bold')
                axs[axi].set_box_aspect(1)
                
                if plot_exist and (plot_exist_style == "both" or plot_exist_style == 'rug'):
                    train_unq_counts[train_unq_counts > 0] = 0.1
                    x_counts = np.ndarray.flatten(train_unq[:,[cvar ==tvars[0] for cvar in cont_vars]])
                    y_counts = np.ndarray.flatten(train_unq[:,[cvar ==tvars[1] for cvar in cont_vars]])
                    tot_range = np.abs(max(np.max(x_counts),np.max(y_counts)) - min(np.min(x_counts),np.min(y_counts)))
                    axs[axi].scatter(x_counts,y_counts,alpha=train_unq_counts,color='black',s=tot_range/(len(x_counts)))

                # Credit to Lasse: https://stackoverflow.com/questions/63118710/
                # This made sure that the colorbar height always matches those of the contour plots.
                axins = inset_axes(axs[axi], width = "5%", height = "100%", loc = 'lower left',
                        bbox_to_anchor = (1.02, 0., 1, 1), bbox_transform = axs[axi].transAxes,
                        borderpad = 0)
                
                if use_ci:
                    cbar = plt.colorbar(axs[axi].collections[1],cax=axins)
                else:
                    cbar = plt.colorbar(axs[axi].collections[0],cax=axins)

                cbar_label = '(' + tvars[0] + ',' + tvars[1] + ')$'

                cbar_label = '$f' + cbar_label

                cbar.ax.set_ylabel(cbar_label,math_fontfamily=math_font,size=math_font_size)

            axi += 1

        # Now handle by terms - essentially we need to perform the above separately for every level
        # of the by/binary factor.
        elif not terms[sti].by is None or not terms[sti].binary is None:
            
            if not terms[sti].by is None:
                sti_by = terms[sti].by
            else:
                sti_by = terms[sti].binary[0]

            levels = list(code_factors[sti_by].keys())

            if not terms[sti].binary is None:
                levels = [factor_codes[sti_by][terms[sti].binary[1]]]

            # Select a small set of levels for random smooths
            if isinstance(terms[sti],fs) and len(levels) > 25:
                levels = np.random.choice(levels,replace=False,size=25)

            if prov_cols is None:
                level_cols = np.linspace(0.1,0.9,len(levels))
            else:
                level_cols = prov_cols

            for level_col,leveli in zip(level_cols,levels):
                pred_level_dat = copy.deepcopy(pred_dat)

                for vari in varmap.keys():
                    if vari in terms[sti].variables:
                        continue
                    else:
                        # Note, placeholder selection must exlcude by/binary variable for which we need to provide the
                        # current level!
                        if vartypes[vari] == VarType.FACTOR and vari == sti_by:
                            if vari in model.formula.get_subgroup_variables():
                                pred_level_dat[vari.split(":")[0]] = [code_factors[vari][leveli] for _ in range(len(x1_exp))]
                            else:
                                pred_level_dat[vari] = [code_factors[vari][leveli] for _ in range(len(x1_exp))]
                        elif vartypes[vari] == VarType.FACTOR:
                            if vari in model.formula.get_subgroup_variables():
                                if sti_by in model.formula.get_subgroup_variables() and sti_by.split(":")[0] == vari.split(":")[0]:
                                    continue
                                
                                pred_level_dat[vari.split(":")[0]] = [code_factors[vari][0] for _ in range(len(x1_exp))]
                            else:
                                pred_level_dat[vari] = [code_factors[vari][0] for _ in range(len(x1_exp))]
                        else:
                            pred_level_dat[vari] = [0 for _ in range(len(x1_exp))]
                
                pred_dat_pd = pd.DataFrame(pred_level_dat)

                # CI-decision - exclude factor smooths if not requested explicitly.
                use_ci = ci
                if use_ci is None:
                    if not isinstance(terms[sti],fs):
                        use_ci = True
                    else:
                        use_ci = False
                
                # Again, add intercept
                use = [sti]
                if use_inter:
                    use = [0,sti]

                if isinstance(model,GAMMLSS):
                    # Reset formula to prevent any problems with the call to predict, since the GAMMLSS class might
                    # change this attribute itself.
                    model.formula = None
                    
                    pred,_,b= model.predict(dist_par,use,pred_dat_pd,ci=use_ci,alpha=ci_alpha,whole_interval=whole_interval,n_ps=n_ps,seed=seed)

                    # Set formula again.
                    model.formula = model.formulas[dist_par]
                else:
                    pred,_,b= model.predict(use,pred_dat_pd,ci=use_ci,alpha=ci_alpha,whole_interval=whole_interval,n_ps=n_ps,seed=seed)

                # Subtract intercept
                if use_inter:
                    _cf,_ = model.get_pars()
                    pred -= _cf[0]

                # Compute data-limits and prepare rug plots
                plot_in_limits = None
                if plot_exist:
                    pred_in_limits,train_unq,train_unq_counts,cont_vars = __get_data_limit_counts(model.formula,pred_dat_pd,tvars,[sti_by])

                if plot_exist and (plot_exist_style == "both" or plot_exist_style == "hide"):
                    plot_in_limits = pred_in_limits

                # Get correct link to transform to response scale
                link = None
                if response_scale:
                    if isinstance(model,GAMMLSS):
                        link = model.family.links[dist_par]
                    else:
                        link = model.family.link

                __pred_plot(pred,b,tvars,plot_in_limits,x1,x2,x1_exp,use_ci,n_vals,axs[axi],_cmp,level_col,ylim,link,None)
                
                # And set up labels again + rug plots if requested
                if not isinstance(terms[sti],fs):

                    if len(tvars) == 1:
                        ax_label = '$f_{' + str(code_factors[sti_by][leveli]) + '}' + '(' + tvars[0] + ')$'
                        axs[axi].set_ylabel(ax_label,math_fontfamily=math_font,size=math_font_size,fontweight='bold')
                        axs[axi].set_xlabel(tvars[0],fontweight='bold')
                        axs[axi].spines['top'].set_visible(False)
                        axs[axi].spines['right'].set_visible(False)
                        
                        if plot_exist and (plot_exist_style == "both" or plot_exist_style == 'rug'):
                    
                            #train_unq_counts /= np.max(train_unq_counts)
                            #train_unq_counts[train_unq_counts > 0] = 1 
                            #pred_range = np.abs(np.max(pred) - np.min(pred))*0.01
                            x_counts = np.ndarray.flatten(train_unq[:,[cvar ==tvars[0] for cvar in cont_vars]])
                            #x_range = np.abs(np.max(x_counts) - np.min(x_counts))

                            #axs[axi].bar(x=x_counts,bottom=axs[axi].get_ylim()[0],height=pred_range*train_unq_counts,color='black',width=max(0.05,x_range/(2*len(x_counts))))
                            axs[axi].scatter(x_counts,[axs[axi].get_ylim()[0]]*len(x_counts),marker='|',color='black',linewidths=0.25)


                    elif len(tvars) == 2:
                        axs[axi].set_ylabel(tvars[1],fontweight='bold')
                        axs[axi].set_xlabel(tvars[0],fontweight='bold')
                        axs[axi].set_box_aspect(1)

                        if plot_exist and (plot_exist_style == "both" or plot_exist_style == 'rug'):

                            train_unq_counts[train_unq_counts > 0] = 0.1
                            x_counts = np.ndarray.flatten(train_unq[:,[cvar ==tvars[0] for cvar in cont_vars]])
                            y_counts = np.ndarray.flatten(train_unq[:,[cvar ==tvars[1] for cvar in cont_vars]])
                            tot_range = np.abs(max(np.max(x_counts),np.max(y_counts)) - min(np.min(x_counts),np.min(y_counts)))
                            axs[axi].scatter(x_counts,y_counts,alpha=train_unq_counts,color='black',s=tot_range/(len(x_counts)))

                        # Credit to Lasse: https://stackoverflow.com/questions/63118710/
                        # This made sure that the colorbar height always matches those of the contour plots.
                        axins = inset_axes(axs[axi], width = "5%", height = "100%", loc = 'lower left',
                                bbox_to_anchor = (1.02, 0., 1, 1), bbox_transform = axs[axi].transAxes,
                                borderpad = 0)
                        
                        if use_ci:
                            cbar = plt.colorbar(axs[axi].collections[1],cax=axins)
                        else:
                            cbar = plt.colorbar(axs[axi].collections[0],cax=axins)

                        cbar_label = '(' + tvars[0] + ',' + tvars[1] + ')$'

                        cbar_label = '$f_{' + str(code_factors[sti_by][leveli]) + '}' + cbar_label

                        cbar.ax.set_ylabel(cbar_label,math_fontfamily=math_font,size=math_font_size)
                    axi += 1

            # Random smooths are all plotted to single figure, so handle labels here. No reason to plot rug
            if isinstance(terms[sti],fs):
                axs[axi].set_ylabel('$f_{' + str(sti_by) + '}(' + tvars[0] + ')$',math_fontfamily=math_font,size=math_font_size,fontweight='bold')
                axs[axi].set_xlabel(tvars[0],fontweight='bold')
                axs[axi].spines['top'].set_visible(False)
                axs[axi].spines['right'].set_visible(False)
                axi += 1
    
    if isinstance(model,GAMMLSS):
        # Clean up
        model.formula = None

    if figs is not None:
        plt.show()


def plot_fitted(pred_dat,tvars,model:GAMM or GAMMLSS,use:[int] or None = None,pred_factors: [str] or None =None,dist_par=0,
                ci=True,ci_alpha=0.05,whole_interval=False,n_ps=10000,seed=None,
                cmp:str or None = None,plot_exist=True,plot_exist_style='both',
                response_scale=True,ax=None,fig_size=(6/2.54,6/2.54),ylim=None,col=0.7,
                label=None,legend_label=False,title=None):
    """Plots the model prediction based on (a subset of) the terms included in the model for new data `pred_dat`.

    In contrast to `plot`, the predictions are by default transformed to the scale of the mean (i.e., response-scale). If `use=None`, the model
    will simply use all parametric and regular smooth terms (but no random smooths) for the prediction (i.e., only the "fixed" effects in the model).

    For a GAMM, a simple example of this function would be::

        # Fit model
        model = GAMM(Formula(lhs("y"),[i(),f(["time"])],data=dat),Gaussian())
        model.fit()

        # Create prediction data
        pred_dat = pd.DataFrame({"time":np.linspace(0,np.max(dat["time"]),30)})

        # Plot predicted mean = \alpha + f(time)
        plot_fitted(pred_dat,["time"],model)

        # This is in contrast to `plot`, which would just visualize pred = f(time)
        plot(model)

    Note that, for predictions visualized as a function of two variables, areas of the prediction for which the CI contains zero will again be visualized with low opacity
    if the CI is to be visualized.
    
    References:

    - Wood, S. N. (2017). Generalized Additive Models: An Introduction with R, Second Edition (2nd ed.).
    - Simpson, G. (2016). Simultaneous intervals for smooths revisited.

    :param pred_dat: A pandas DataFrame containing new data for which to make the prediction. Importantly, all variables present in the data used to fit the model also need to be present in this DataFrame. Additionally, factor variables must only include levels
    also present in the data used to fit the model. If you want to exclude a specific factor from the prediction (for example the factor subject) don't include the terms that involve it in the ``use`` argument.
    :type pred_dat: pandas.DataFrame
    :param tvars: List of variables to be visualized - must contain one string for predictions visualized as a function of a single variable, two for predictions visualized as a function of two variables
    :type tvars: [str]
    :param model: The estimated GAMM or GAMMLSS model for which the visualizations are to be obtained
    :type model: GAMM or GAMMLSS
    :param use: The indices corresponding to the terms that should be used to obtain the prediction or ``None`` in which case all fixed effects will be used, defaults to None
    :type use: [int] or None, optional
    :param pred_factors: List of factor variables to consider for data limit/availability computations - by default, all factor variables in the model are considered.
    :type pred_factors: [str]
    :param dist_par: The index corresponding to the parameter for which to make the prediction (e.g., 0 = mean) - only necessary if a GAMMLSS model is provided, defaults to 0
    :type dist_par: int, optional
    :param ci: Whether the standard error ``se`` for credible interval (CI; see  Wood, 2017) calculation should be computed and used to visualize CIs. The CI is then [``pred`` - ``se``, ``pred`` + ``se``], defaults to None
    in which case the CI will be visualized for fixed effects but not for random smooths
    :type ci: bool or None, optional
    :param ci_alpha: The alpha level to use for the standard error calculation. Specifically, 1 - (``alpha``/2) will be used to determine the critical cut-off value according to a N(0,1), defaults to 0.05
    :type ci_alpha: float, optional
    :param whole_interval: Whether or not to adjuste the point-wise CI to behave like whole-interval (based on Wood, 2017; section 6.10.2 and Simpson, 2016), defaults to False
    :type whole_interval: bool, optional
    :param n_ps: How many samples to draw from the posterior in case the point-wise CI is adjusted to behave like a whole-interval CI, defaults to 10000
    :type n_ps: int, optional
    :param seed: Can be used to provide a seed for the posterior sampling step in case the point-wise CI is adjusted to behave like a whole-interval CI, defaults to None
    :type seed: int, optional
    :param cmp: string corresponding to name for a matplotlib colormap, defaults to None in which case it will be set to 'RdYlBu_r'.
    :type cmp: str or None, optional
    :param plot_exist: Whether or not an indication of the data distribution should be provided. For predictions visualized as a function of a single variable setting this to True will add a rug-plot to the
    bottom, indicating for which covariate values samples existed in the training data. For predictions visualized as a function of a two variables setting this to true will result in a 2d scatter rug plot being added and/or values outside of data limits being hidden, defaults to True
    :type plot_exist: bool, optional
    :param plot_exist_style: Determines the style of the data distribution indication for smooths. Must be 'rug', 'hide',or 'both'. 'both' will both add the rug-plot and hide values out of data limits, defaults to 'both'
    :type plot_exist_style: str, optional
    :param response_scale: Whether or not predictions and CIs should be shown on the scale of the model predictions (linear scale) or on the 'response-scale' i.e., the scale of the mean, defaults to True
    :type response_scale: bool, optional
    :param ax: A matplotlib.axis on which the Figure should be drawn, defaults to None in which case an axis will be created by the function and plot.show() will be called at the end
    :type ax: matplotlib.axis, optional
    :param fig_size: Tuple holding figure size, which will be used to determine the size of the figures created if `ax=None`, defaults to (6/2.54,6/2.54)
    :type fig_size: tuple, optional
    :param ylim: Tuple holding y-limits, defaults to None in which case y_limits will be inferred from the predictions made
    :type ylim: (float,float), optional
    :param col: A float in [0,1]. Used to get a color for univariate predictions from the chosen colormap, defaults to 0.7
    :type col: float, optional
    :param label: A labels to add to the y axis for univariate predictions (or to a legend, if `legend_label`=True) or to the color-bar for tensor predictions, defaults to None
    :type label: str, optional
    :param legend_label: Whether or not any `label` should be added to a legend (don't forget to call :func:`plt.legend()`) or to the y-axis for univariate predicitions, defaults to False (the latter)
    :type legend_label: bool, optional
    :param title: A list of titles to add to each plot, defaults to None
    :type title: [str], optional
    :raises ValueError: If a visualization is requested for more than 2 variables
    """
    
    if isinstance(model,GAMMLSS):
        # Set up everything for that we can plot all smooth terms for distribution parameter `dist_par`.
        model.formula = model.formulas[dist_par]
    
    # Select only fixed effects if nothing is provided
    if use is None:
        use = model.formula.get_linear_term_idx()

        terms = model.formula.get_terms()
        for sti in model.formula.get_smooth_term_idx():
            if not isinstance(terms[sti],fs):
                use.append(sti)
    
    # Create figure if necessary
    fig = None
    if ax is None:
        fig = plt.figure(figsize=fig_size,layout='constrained')
        ax = fig.add_subplot(1,1,1)
    
    # Set up predictor variables as done in `plot`
    x1_exp = np.array(pred_dat[tvars[0]])
    x1 = np.unique(x1_exp)
    x2 = None
    if len(tvars) == 2:
        x2 = np.unique(pred_dat[tvars[1]])

    elif len(tvars) > 2:
        raise ValueError("Can only visualize fitted effects over one or two continuous variables.")
    
    if cmp is None:
        cmp = 'RdYlBu_r'
        
    _cmp = matplotlib.colormaps[cmp] 

    if isinstance(model,GAMMLSS):
        # Reset formula to prevent any problems with the call to predict, since the GAMMLSS class might
        # change this attribute itself.
        model.formula = None
        
        pred,_,b= model.predict(dist_par,use,pred_dat,ci=ci,alpha=ci_alpha,whole_interval=whole_interval,n_ps=n_ps,seed=seed)

        # Set formula again.
        model.formula = model.formulas[dist_par]
    else:
        pred,_,b= model.predict(use,pred_dat,ci=ci,alpha=ci_alpha,whole_interval=whole_interval,n_ps=n_ps,seed=seed)

    # Optionally get data limits
    plot_in_limits = None
    if plot_exist:
        if pred_factors is None:
            pred_factors = [var for var in pred_dat.columns if model.formula.get_var_types()[var] == VarType.FACTOR]
        if (not pred_factors is None) and len(pred_factors) == 0:
            pred_factors = None

        pred_in_limits,train_unq,train_unq_counts,cont_vars = __get_data_limit_counts(model.formula,pred_dat,tvars,pred_factors)

    if plot_exist and (plot_exist_style == "both" or plot_exist_style == "hide"):
        plot_in_limits = pred_in_limits

    # By default transform predictions to scale of mean
    link = None
    if response_scale:
        if isinstance(model,GAMMLSS):
            link = model.family.links[dist_par]
        else:
            link = model.family.link
    
    plot_label = None
    if legend_label and len(tvars) == 1:
        plot_label = label

    __pred_plot(pred,b,tvars,plot_in_limits,x1,x2,x1_exp,ci,len(x1),ax,_cmp,col,ylim,link,plot_label)

    # Label axes + visualize rug plots if requested
    if len(tvars) == 2:
       
        if plot_exist and (plot_exist_style == "both" or plot_exist_style == 'rug'):

            train_unq_counts[train_unq_counts > 0] = 0.1
            x_counts = np.ndarray.flatten(train_unq[:,[cvar ==tvars[0] for cvar in cont_vars]])
            y_counts = np.ndarray.flatten(train_unq[:,[cvar ==tvars[1] for cvar in cont_vars]])
            tot_range = np.abs(max(np.max(x_counts),np.max(y_counts)) - min(np.min(x_counts),np.min(y_counts)))
            ax.scatter(x_counts,y_counts,alpha=train_unq_counts,color='black',s=tot_range/(len(x_counts)))

        # Credit to Lasse: https://stackoverflow.com/questions/63118710/
        # This made sure that the colorbar height always matches those of the contour plots.
        axins = inset_axes(ax, width = "5%", height = "100%", loc = 'lower left',
                bbox_to_anchor = (1.02, 0., 1, 1), bbox_transform = ax.transAxes,
                borderpad = 0)
        
        if ci:
            cbar = plt.colorbar(ax.collections[1],cax=axins)
        else:
            cbar = plt.colorbar(ax.collections[0],cax=axins)

        if not label is None:
            cbar.set_label(label,fontweight='bold')
        else:
            cbar.set_label("Predicted",fontweight='bold')

        ax.set_xlabel(tvars[0],fontweight='bold')
        ax.set_ylabel(tvars[1],fontweight='bold')
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
    else:
        if not label is None and (legend_label==False):
            ax.set_ylabel(label,fontweight='bold')
        else:
            ax.set_ylabel("Predicted",fontweight='bold')
        ax.set_xlabel(tvars[0],fontweight='bold')
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)

        if plot_exist and (plot_exist_style == "both" or plot_exist_style == 'rug'):
                    
            #train_unq_counts[train_unq_counts > 0] = 1 
            #pred_range = np.abs(np.max(pred) - np.min(pred))*0.025
            x_counts = np.ndarray.flatten(train_unq[:,[cvar ==tvars[0] for cvar in cont_vars]])
            #x_range = np.abs(np.max(x_counts) - np.min(x_counts))

            ax.scatter(x_counts,[ax.get_ylim()[0]]*len(x_counts),marker='|',color='black',linewidths=0.25)
    
    if not title is None:
        ax.set_title(title,fontweight='bold')

    if isinstance(model,GAMMLSS):
        # Clean up
        model.formula = None

    if fig is not None:
        plt.show()

def plot_diff(pred_dat1,pred_dat2,tvars,model: GAMM or GAMMLSS,use:[int] or None = None,dist_par=0,
              ci_alpha=0.05,whole_interval=False,n_ps=10000,seed=None,cmp:str or None = None,
              plot_exist=True,response_scale=True,ax=None,n_vals=30,fig_size=(6/2.54,6/2.54),
              ylim=None,col=0.7,label=None,title=None):
    """Plots the expected difference (and CI around this expected difference) between two sets of predictions, evaluated for `pred_dat1` and `pred_dat2`.

    This function is primarily designed to visualize the expected difference between two levels of a categorical/factor variable. For example, consider the following
    model below, including a separate smooth of "time" per level of the factor "cond". It is often of interest to visualize *when* in time the two levels of "cond" differ
    from each other in their dependent variable. For this, the difference curve over "time", essentially the smooth of "time" for the first level subtracted from the
    smooth of "time" for the second level of factor "cond" (offset terms can also be accounted for, check the `use` argument), can be visualized together with a CI (Wood, 2017).
    This CI can provide insights into whether and *when* the two levels can be expected to be different from each other. To visualize this difference curve as well as the
    difference CI, this function can be used as follows::

        # Define & estimate model
        model = GAMM(Formula(lhs("y"),[i(), l(["cond"]), f(["time"],by="cond")],data=dat),Gaussian())
        model.fit()

        # Create prediction data, differing only in the level of factor cond
        time_pred = np.linspace(0,np.max(dat["time"]),30)
        new_dat1 = pd.DataFrame({"cond":["a" for _ in range(len(time_pred))],
                                "time":time_pred})

        new_dat2 = pd.DataFrame({"cond":["b" for _ in range(len(time_pred))],
                                "time":time_pred})

        # Now visualize diff = (\alpha_a + f_a(time)) - (\alpha_b + f_b(time)) and the CI around diff
        plot_diff(pred_dat1,pred_dat2,["time"],model)

    This is only the most basic example to illustrate the usefulness of this function. Many other options are possible. Consider for example the model below, which allows for
    the expected time-course to vary smoothly as a function of additional covariate "x" - achieved by inclusion of the tensor smooth term of "time" and "x". In addition, this
    model allows for the shape of the tensor smooth to differ between the levels of factor "cond"::

        model = GAMM(Formula(lhs("y"),[i(), l(["cond"]), f(["time","x"],by="cond",te=True)],data=dat),Gaussian())

    For such a model, multiple predicted differences might be of interest. One option would be to look only at a single level of "cond" and to visualize the predicted difference
    in the time-course for two different values of "x" (perhaps two quantiles). In that case, `pred_dat1` and `pred_dat2` would have to be set up to differ only in the value of
    "x" - they should be equivalent in terms of "time" and "cond" values.

    Alternatively, it might be of interest to look at the predicted difference between the tensor smooth surfaces for two levels of factor "cond". Rather than being interested in
    a difference curve, this would mean we are interested in a difference *surface*. To achieve this, `pred_dat1` and `pred_dat2` would again have to be set up to differ only in the
    value of "cond" - they should be equivalent in terms of "time" and "x" values. In addition, it would be necessary to specify `tvars=["time","x"]`. Note that, for such difference surfaces,
    areas of the difference prediction for which the CI contains zero will again be visualized with low opacity if the CI is to be visualized.
    
    References:

    - Wood, S. N. (2017). Generalized Additive Models: An Introduction with R, Second Edition (2nd ed.).
    - Simpson, G. (2016). Simultaneous intervals for smooths revisited.

    :param pred_dat1: A pandas DataFrame containing new data for which the prediction is to be compared to the prediction obtained for `pred_dat2`. Importantly, all variables present in the data used to fit the model also need to be present in this DataFrame. Additionally, factor variables must only include levels
    also present in the data used to fit the model. If you want to exclude a specific factor from the difference prediction (for example the factor subject) don't include the terms that involve it in the ``use`` argument.
    :type pred_dat1: pandas.DataFrame
    :param pred_dat2: Like `pred_dat1` - ideally differing only in the level of a single factor variable or the value of a single continuous variable.
    :type pred_dat2: pandas.DataFrame
    :param tvars: List of variables to be visualized - must contain one string for difference predictions visualized as a function of a single variable, two for difference predictions visualized as a function of two variables
    :type tvars: [str]
    :param model: The estimated GAMM or GAMMLSS model for which the visualizations are to be obtained
    :type model: GAMM or GAMMLSS
    :param use: The indices corresponding to the terms that should be used to obtain the prediction or ``None`` in which case all fixed effects will be used, defaults to None
    :type use: [int] or None, optional
    :param dist_par: The index corresponding to the parameter for which to make the prediction (e.g., 0 = mean) - only necessary if a GAMMLSS model is provided, defaults to 0
    :type dist_par: int, optional
    :param ci_alpha: The alpha level to use for the standard error calculation. Specifically, 1 - (``alpha``/2) will be used to determine the critical cut-off value according to a N(0,1), defaults to 0.05
    :type ci_alpha: float, optional
    :param whole_interval: Whether or not to adjuste the point-wise CI to behave like whole-interval (based on Wood, 2017; section 6.10.2 and Simpson, 2016), defaults to False
    :type whole_interval: bool, optional
    :param n_ps: How many samples to draw from the posterior in case the point-wise CI is adjusted to behave like a whole-interval CI, defaults to 10000
    :type n_ps: int, optional
    :param seed: Can be used to provide a seed for the posterior sampling step in case the point-wise CI is adjusted to behave like a whole-interval CI, defaults to None
    :type seed: int, optional
    :param cmp: string corresponding to name for a matplotlib colormap, defaults to None in which case it will be set to 'RdYlBu_r'.
    :type cmp: str or None, optional
    :param plot_exist: Whether or not an indication of the data distribution should be provided. For difference predictions visualized as a function of a single variable this will simply hide predictions outside of the data-limits. For difference predictions visualized as a function of a two variables setting this to true will result in values outside of data limits being hidden, defaults to True
    :type plot_exist: bool, optional
    :param response_scale: Whether or not predictions and CIs should be shown on the scale of the model predictions (linear scale) or on the 'response-scale' i.e., the scale of the mean, defaults to True
    :type response_scale: bool, optional
    :param ax: A matplotlib.axis on which the Figure should be drawn, defaults to None in which case an axis will be created by the function and plot.show() will be called at the end
    :type ax: matplotlib.axis, optional
    :param fig_size: Tuple holding figure size, which will be used to determine the size of the figures created if `ax=None`, defaults to (6/2.54,6/2.54)
    :type fig_size: tuple, optional
    :param ylim: Tuple holding y-limits, defaults to None in which case y_limits will be inferred from the predictions made
    :type ylim: (float,float), optional
    :param col: A float in [0,1]. Used to get a color for univariate predictions from the chosen colormap, defaults to 0.7
    :type col: float, optional
    :param label: A list of labels to add to the y axis for univariate predictions or to the color-bar for tensor predictions, defaults to None
    :type label: [str], optional
    :param title: A list of titles to add to each plot, defaults to None
    :type title: [str], optional
    :raises ValueError: If a visualization is requested for more than 2 variables
    """
    
    if isinstance(model,GAMMLSS):
        # Set up everything for that we can plot all smooth terms for distribution parameter `dist_par`.
        model.formula = model.formulas[dist_par]

    if use is None:
        use = model.formula.get_linear_term_idx()

        terms = model.formula.get_terms()
        for sti in model.formula.get_smooth_term_idx():
            if not isinstance(terms[sti],fs):
                use.append(sti)

    fig = None
    if ax is None:
        fig = plt.figure(figsize=fig_size,layout='constrained')
        ax = fig.add_subplot(1,1,1)
    
    x1_exp = np.array(pred_dat1[tvars[0]])
    x1 = np.unique(x1_exp)
    x2 = None
    if len(tvars) == 2:
        x2 = np.unique(pred_dat1[tvars[1]])

    elif len(tvars) > 2:
        raise ValueError("Can only visualize fitted effects over one or two continuous variables.")
    
    if cmp is None:
        cmp = 'RdYlBu_r'
        
    _cmp = matplotlib.colormaps[cmp] 


    if isinstance(model,GAMMLSS):
        # Reset formula to prevent any problems with the call to predict, since the GAMMLSS class might
        # change this attribute itself.
        model.formula = None
        
        pred,_,b= model.predict_diff(pred_dat1,pred_dat2,dist_par,use,alpha=ci_alpha,whole_interval=whole_interval,n_ps=n_ps,seed=seed)

        # Set formula again.
        model.formula = model.formulas[dist_par]
    else:
        pred,b= model.predict_diff(pred_dat1,pred_dat2,use,alpha=ci_alpha,whole_interval=whole_interval,n_ps=n_ps,seed=seed)

    in_limits = None
    if plot_exist:
        pred_factors1 = [var for var in pred_dat1.columns if model.formula.get_var_types()[var] == VarType.FACTOR]
        if len(pred_factors1) == 0:
            pred_factors1 = None
        pred_in_limits1,train_unq1,train_unq_counts1,cont_vars1 = __get_data_limit_counts(model.formula,pred_dat1,tvars,pred_factors1)

        pred_factors2 = [var for var in pred_dat2.columns if model.formula.get_var_types()[var] == VarType.FACTOR]
        if len(pred_factors2) == 0:
            pred_factors2 = None
        pred_in_limits2,train_unq2,train_unq_counts2,cont_vars2 = __get_data_limit_counts(model.formula,pred_dat2,tvars,pred_factors2)

        in_limits = pred_in_limits1 & pred_in_limits2
    
    link = None
    if response_scale:
        if isinstance(model,GAMMLSS):
            link = model.family.links[dist_par]
        else:
            link = model.family.link

    __pred_plot(pred,b,tvars,in_limits,x1,x2,x1_exp,True,n_vals,ax,_cmp,col,ylim,link,None)

    if len(tvars) == 2:
        # Credit to Lasse: https://stackoverflow.com/questions/63118710/
        # This made sure that the colorbar height always matches those of the contour plots.
        axins = inset_axes(ax, width = "5%", height = "100%", loc = 'lower left',
                bbox_to_anchor = (1.02, 0., 1, 1), bbox_transform = ax.transAxes,
                borderpad = 0)
        
        cbar = plt.colorbar(ax.collections[1],cax=axins)

        if not label is None:
            cbar.set_label(label,fontweight='bold')
        else:
            cbar.set_label("Predicted Difference",fontweight='bold')
        
        ax.set_xlabel(tvars[0],fontweight='bold')
        ax.set_ylabel(tvars[1],fontweight='bold')
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
    else:
        if not label is None:
            ax.set_ylabel(label,fontweight='bold')
        else:
            ax.set_ylabel("Predicted Difference",fontweight='bold')
        ax.set_xlabel(tvars[0],fontweight='bold')
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)

        if plot_exist:
            ax.set_xlim(min(x1),max(x1))
    
    if not title is None:
        ax.set_title(title,fontweight='bold')
    
    if isinstance(model,GAMMLSS):
        # Clean up
        model.formula = None

    if fig is not None:
        plt.show()


def plot_val(model:GAMM or GAMMLSS,pred_viz:[str] or None = None,resid_type="deviance",
             ar_lag=100,response_scale=False,qq=True,axs=None,fig_size=(6/2.54,6/2.54)):
    """Plots residual plots useful for validating whether the `model` meets the regression assumptions.

    At least four plots will be generated:

    - A scatter-plot: Model predictions (always on response/mean scale) vs. Observations
    - A scatter-plot: Model predictions (optionally on response/mean scale) vs. Residuals
    - A Histogram/QQ-plot: Residuals (with density overlay of expected distribution)/Quantile-quantile plot for residuals against theoretical quantiles.
    - An ACF plot: Showing the auto-correlation in the residuals at each of `ar_lag` lags
    
    For each additional predictor included in `pred_viz`, an additional scatter-plot will be generated plotting the predictor values against the residuals.

    Which residuals will be visualized depends on the choice of `model` and `resid_type`. If `model` is a `GAMM` model, `resid_type` will determine whether
    "Pearson" or "Deviance" (default) residuals are to be plotted (Wood, 2017). Except, for a Gaussian `GAMM`, in which case the function will always plot
    the classical residuals, i.e., the difference between model predictions (\mu_i) and observed values (y_i). This ensures that by default and for any `GAMM`
    we can expect the residuals to look like N independent samples from N(0,sqrt(phi)) - where \phi is the scale parameter of the `GAMM` (\sigma^2 for Gaussian).
    Hence, we can interpret all the plots in the same way. Note, that residuals for Binomial models will generally not look pretty or like N(0,sqrt(phi)) - but they
    should still be reasonably independent.

    If `model` is a `GAMMLSS` model, `resid_type` will be ignored. Instead, the function will always plot standardized residuals that behave a lot like deviance residuals, except for
    the fact that we also cancel for \phi, so that we can expect the residuals to look like N independent samples from N(0,1). In many cases, the computation of these residuals will
    thus follow the computation for GAMMs (for a Gaussian GAMMLSS model we can for example simply scale the residual vector by the sigma parameter estimated for each observation to achieve
    the desired distribution result) while for others more complicated standardization might be necessary (see Rigby & Stasinopoulos, 2005) - this will be noted in the docstring of
    the :func:`resid()` method implemented by each `GAMLSSFamily` family. Again, we have to make an exeception for the Multinomial model (`MULNOMLSS`), which is currently not supported by this function.

    References:

    - Rigby, R. A., & Stasinopoulos, D. M. (2005). Generalized Additive Models for Location, Scale and Shape.
    - Wood, S. N. (2017). Generalized Additive Models: An Introduction with R, Second Edition (2nd ed.).

    :param model: Estimated GAMM or GAMMLSS model, for which the reisdual plots should be generated.
    :type model: GAMM or GAMMLSS
    :param pred_viz: A list of additional predictor variables included in the model. For each one provided an additional plot will be created with the predictor on the x-axis and the residuals on the y-axis, defaults to None
    :type pred_viz: [str] or None, optional
    :param resid_type: Type of residual to visualize. For a `model` that is a GAMM this can be "Pearson" or "Deviance" - for a Gaussian GAMM the function will alwasy plot default residuals (y_i - \mu_i) independent of what is provided.
    For a `model` that is a GAMMLSS, the function will always plot standardized residuals that should approximately behave like deviance ones - except that they can be expected to look like N(0,1) if the model is specified correctly, defaults to "deviance"
    :type resid_type: str, optional
    :param ar_lag: Up to which lag the auto-correlation function in the residuals should be computed and visualized, defaults to 100
    :type ar_lag: int, optional
    :param response_scale: Whether or not predictions should be visualized on the scale of the mean or not, defaults to False - i.e., predictions are visualized on the scale of the model predictions/linear scale
    :type response_scale: bool, optional
    :param qq: Whether or not a qq-plot should be drawn instead of a Histogram, defaults to True
    :type qq: bool, optional
    :param axs: A list of matplotlib.axis on which Figures should be drawn, defaults to None in which case axis will be created by the function and plot.show() will be called at the end
    :type axs: [matplotlib.axis], optional
    :param fig_size: Tuple holding figure size, which will be used to determine the size of the figures created if `axs=None`, defaults to (6/2.54,6/2.54)
    :type fig_size: tuple, optional
    :raises ValueError: If fewer matplotlib axis are provided than the number of figures that would be created
    :raises TypeError: If the function is called with a `model` of the `MULNOMLSS` family, which is currently not supported
    """

    if isinstance(model.family,MULNOMLSS):
        raise TypeError("Function does not currently support `Multinomial` models.")

    if isinstance(model,GAMMLSS):
        # Set up everything for that we can plot all smooth terms for distribution parameter `dist_par`.
        model.formula = model.formulas[0]

    varmap = model.formula.get_var_map()
    n_figures = 4

    if pred_viz is not None:
        for pr in pred_viz:
            n_figures +=1
    
    if not axs is None and len(axs) != n_figures:
        raise ValueError(f"{n_figures} plots would be created, but only {len(axs)} axes were provided!")

    figs = None
    if axs is None:
        figs = [plt.figure(figsize=fig_size,layout='constrained') for _ in range(n_figures)]
        axs = [fig.add_subplot(1,1,1) for fig in figs]
    
    if isinstance(model,GAMMLSS) == False:
        _, sigma = model.get_pars() # sigma = **variance** of residuals!
        pred = model.pred # The model prediction for the entire data
    else:
        sigma = 1 # Standardized residuals should look like N(0,1)
        pred = model.overall_preds[0]

    if response_scale:
        if isinstance(model,GAMMLSS):
            pred = model.family.links[0].fi(pred)
        else:
            pred = model.family.link.fi(pred)

    if isinstance(model,GAMMLSS) == False:
        res = model.get_resid(type=resid_type)
    else:
        res = model.get_resid() # resid are alwasy standardized for GAMMLSS models

    y = model.formula.y_flat[model.formula.NOT_NA_flat] # The dependent variable after NAs were removed

    # obs vs. pred plot should always be on response scale
    if isinstance(model,GAMMLSS) == False and (response_scale == False):
        axs[0].scatter(model.family.link.fi(pred),y,color="black",facecolor='none')
    elif isinstance(model,GAMMLSS) and (response_scale == False):
        axs[0].scatter(model.family.links[0].fi(pred),y,color="black",facecolor='none')
    else:
        axs[0].scatter(pred,y,color="black",facecolor='none')

    axs[0].set_xlabel("Predicted (Mean scale)",fontweight='bold')
    axs[0].set_ylabel("Observed",fontweight='bold')
    axs[0].spines['top'].set_visible(False)
    axs[0].spines['right'].set_visible(False)

    axs[1].scatter(pred,res,color="black",facecolor='none')
    if response_scale == False:
        axs[1].set_xlabel("Predicted",fontweight='bold')
    else:
        axs[1].set_xlabel("Predicted (Mean scale)",fontweight='bold')
    axs[1].set_ylabel("Residuals",fontweight='bold')
    axs[1].spines['top'].set_visible(False)
    axs[1].spines['right'].set_visible(False)

    axi = 2

    if pred_viz is not None:
        for pr in pred_viz:
            pr_val =  model.formula.cov_flat[model.formula.NOT_NA_flat,varmap[pr]]
            axs[axi].scatter(pr_val,res,color="black",facecolor='none')
            axs[axi].set_xlabel(pr,fontweight='bold')
            axs[axi].set_ylabel("Residuals",fontweight='bold')
            axs[axi].spines['top'].set_visible(False)
            axs[axi].spines['right'].set_visible(False)
            axi += 1

    # Histogram for normality
    if qq==False:
        axs[axi].hist(res,bins=100,density=True,color="black")
        x = np.linspace(scp.stats.norm.ppf(0.0001,scale=math.sqrt(sigma)),
                        scp.stats.norm.ppf(0.9999,scale=math.sqrt(sigma)), 100)

        axs[axi].plot(x, scp.stats.norm.pdf(x,scale=math.sqrt(sigma)),
                'r-', lw=3, alpha=0.6)

        axs[axi].set_xlabel("Residuals",fontweight='bold')
        axs[axi].set_ylabel("Density",fontweight='bold')
        axs[axi].spines['top'].set_visible(False)
        axs[axi].spines['right'].set_visible(False)
    else:
        # Get theoretical quantiles
        tq = scp.stats.norm.ppf(np.linspace(0,1,len(res)),scale=math.sqrt(sigma))
        # Get empirical quantiles
        eq = np.quantile(res,np.linspace(0,1,len(res)))
        axs[axi].scatter(tq,eq,color="black",facecolor='none')
        # Add theoretical truth line
        axs[axi].plot(tq,tq,color="black")

        axs[axi].set_xlabel("Theoretical Quantiles",fontweight='bold')
        axs[axi].set_ylabel("Empirical Residual Quantiles",fontweight='bold')
        axs[axi].spines['top'].set_visible(False)
        axs[axi].spines['right'].set_visible(False)

    axi += 1

    # Auto-correlation check
    cc = np.vstack([res[:-ar_lag,0],*[res[l:-(ar_lag-l),0] for l in range(1,ar_lag)]]).T
    acf = [np.corrcoef(cc[:,0],cc[:,l])[0,1] for l in range(ar_lag)]

    for lg in range(ar_lag):
        axs[axi].plot([lg,lg],[0,acf[lg]],color="black",linewidth=0.5)

    axs[axi].axhline(0,color="red")
    axs[axi].set_xlabel("Lag",fontweight='bold')
    axs[axi].set_ylabel("ACF",fontweight='bold')
    axs[axi].spines['top'].set_visible(False)
    axs[axi].spines['right'].set_visible(False)

    if isinstance(model,GAMMLSS):
        # Clean up
        model.formula = None

    if figs is not None:
        plt.show()